"""
Module for implementation of the :class:`.TensorProductSpace` class and
related methods.
"""
import copy
from numbers import Number
import sympy as sp
import numpy as np
from shenfun import config
from shenfun.fourier.bases import R2C, C2C
from shenfun.utilities import apply_mask
from shenfun.forms.arguments import Function, Array
from shenfun.optimization.cython import evaluate
from shenfun.spectralbase import slicedict, islicedict, SpectralBase
from shenfun.coordinates import Coordinates
from mpi4py_fft.mpifft import Transform, PFFT
from mpi4py_fft.pencil import Subcomm, Pencil
from mpi4py import MPI

comm = MPI.COMM_WORLD

__all__ = ('TensorProductSpace', 'VectorSpace', 'TensorSpace',
           'CompositeSpace', 'Convolve')

@staticmethod
def _get_kind(xfftn, kind):
    try:
        family = xfftn.func.func.__self__.family()
        kind = {} if kind is None else kind
        return kind.get(family, config['transforms']['kind'][family])
    except:
        return None

@staticmethod
def _get_mesh(xfftn, mesh):
    axis = xfftn.func.func.__self__.axis
    if isinstance(mesh, TensorProductSpace):
        return mesh.bases[axis]
    return mesh

# Monkey-patch
Transform._get_kind = _get_kind
Transform._get_mesh = _get_mesh

class BackwardTransform(Transform):
    def __call__(self, input_array=None, output_array=None, kind=None, mesh=None, **kw):
        """Compute backward transform

        Parameters
        ----------
        input_array : array, optional
        output_array : array, optional
        kind : dict, optional
            Alternatives for implementation. Can be used to overload
            config['transforms']['kind'], with a dictionary with family as key
            and values either one of the strings
            - 'fast'
            - 'recursive'
            - 'vandermonde'
            For example kind={'chebyshev': 'vandermonde'}

        mesh : str or functionspace, optional
            - 'quadrature' - use quadrature mesh
            - 'uniform' - use uniform mesh
            - instance of :class:`.TensorProductSpace` or :class:`.CompositeSpace`
              Use Vandermonde and the quadrature mesh of this space.
              The space must have the same underlying mesh distribution as self

        kw : dict
            parameters to serial transforms
            Note in particular that the keyword ``'normalize'=True/False`` can be
            used to turn normalization on or off. Default is to enable
            normalization for forward transforms and disable it for backward.

        Note
        ----
        If input_array/output_array are not given, then use predefined arrays
        as planned with serial transform object ``_xfftn``.

        """

        if input_array is not None:
            self.input_array[...] = input_array

        for i in range(len(self._transfer)):
            self._xfftn[i](kind=self._get_kind(self._xfftn[i], kind), mesh=self._get_mesh(self._xfftn[i], mesh), **kw)
            arrayA = self._xfftn[i].output_array
            arrayB = self._xfftn[i+1].input_array
            self._transfer[i](arrayA, arrayB)
        self._xfftn[-1](kind=self._get_kind(self._xfftn[-1], kind), mesh=self._get_mesh(self._xfftn[-1], mesh), **kw)

        if output_array is not None:
            output_array[...] = self.output_array
            return output_array
        else:
            return self.output_array

class ScalarTransform(Transform):
    """Class for performing the scalar product in parallel

    Parameters
    ----------
    xfftn : list of serial transform objects
    transfer : list of global redistribution objects
    pencil : list of two pencil objects
        The two pencils represent the input and final output configuration of
        the distributed global arrays
    T : :class:`.TensorProductSpace`
    """
    def __init__(self, xfftn, transfer, pencil, T=None):
        assert len(xfftn) == len(transfer) + 1 and len(pencil) == 2
        assert T is not None
        self._xfftn = tuple(xfftn)
        self._transfer = tuple(transfer)
        self._pencil = tuple(pencil)
        self._T = T

    @property
    def sg(self):
        """Return scaling factors of Transform"""
        return self._T.coors.get_sqrt_det_g()

    @property
    def local_mesh(self):
        return self._T.local_mesh(True)

    def get_measured_input_array(self):
        dx = self.sg
        mesh = self.local_mesh
        if dx == 1:
            return

        if isinstance(dx, Number):
            self.input_array[:] = self.input_array*dx
            return

        sym0 = dx.free_symbols
        m = []
        for sym in sym0:
            j = 'xyzrs'.index(str(sym))
            m.append(mesh[j])
        xj = sp.lambdify(tuple(sym0), dx)(*m)
        self.input_array[...] = self.input_array*xj
        return

    def __call__(self, input_array=None, output_array=None, kind=None, **kw):
        """Compute scalar product

        Parameters
        ----------
        input_array : array, optional
        output_array : array, optional
        kind : dict, optional
            Alternatives for implementation. Can be used to overload
            config['transforms']['kind'], with a dictionary with family as key
            and values either one of the strings
            - 'fast'
            - 'recursive'
            - 'vandermonde'
            For example kind={'chebyshev': 'vandermonde'}
        kw : dict
            parameters to serial transforms
            Note in particular that the keyword ``'normalize'=True/False`` can be
            used to turn normalization on or off. Default is to enable
            normalization for forward transforms and disable it for backward.

        Note
        ----
        If input_array/output_array are not given, then use predefined arrays
        as planned with serial transform object ``_xfftn``.

        """
        if input_array is not None:
            self.input_array[...] = input_array

        if not self._T.coors.is_cartesian:
            self.get_measured_input_array()

        for i in range(len(self._transfer)):
            self._xfftn[i](kind=self._get_kind(self._xfftn[i], kind), **kw)
            arrayA = self._xfftn[i].output_array
            arrayB = self._xfftn[i+1].input_array
            self._transfer[i](arrayA, arrayB)
        self._xfftn[-1](kind=self._get_kind(self._xfftn[-1], kind), **kw)

        if output_array is not None:
            output_array[...] = self.output_array
            return output_array
        else:
            return self.output_array


class ForwardTransform(ScalarTransform):
    """Class for performing the forward transform in parallel

    Parameters
    ----------
    xfftn : list of serial transform objects
    transfer : list of global redistribution objects
    pencil : list of two pencil objects
        The two pencils represent the input and final output configuration of
        the distributed global arrays
    T : :class:`.TensorProductSpace`
    """
    def __call__(self, input_array=None, output_array=None, kind=None, **kw):
        """Compute forward transform

        Parameters
        ----------
        input_array : array, optional
        output_array : array, optional
        kind : dict, optional
            Alternatives for implementation. Can be used to overload
            config['transforms']['kind'], with a dictionary with family as key
            and values either one of the strings
            - 'fast'
            - 'recursive'
            - 'vandermonde'
            For example kind={'chebyshev': 'vandermonde'}
        kw : dict
            parameters to serial transforms
            Note in particular that the keyword ``'normalize'=True/False`` can be
            used to turn normalization on or off. Default is to enable
            normalization for forward transforms and disable it for backward.

        Note
        ----
        If input_array/output_array are not given, then use predefined arrays
        as planned with serial transform object ``_xfftn``.

        """
        if input_array is not None:
            self.input_array[...] = input_array

        if not self._T.coors.is_cartesian:
            self.get_measured_input_array()

        if len(self._T.get_nonhomogeneous_axes()) > 1:
            from shenfun import la, TestFunction, TrialFunction, inner
            assert self._T.dimensions == 2, 'Two inhomogeneous boundary directions only implemented for 2D'
            u = TrialFunction(self._T)
            v = TestFunction(self._T)
            B = inner(u, v, kind=kind)
            b = inner(v, input_array, kind=kind)
            sol = la.Solver2D(B)
            self.output_array[:] = sol(b)
        else:
            for i in range(len(self._transfer)):
                self._xfftn[i](kind=self._get_kind(self._xfftn[i], kind), **kw)
                arrayA = self._xfftn[i].output_array
                arrayB = self._xfftn[i+1].input_array
                self._transfer[i](arrayA, arrayB)
            self._xfftn[-1](kind=self._get_kind(self._xfftn[-1], kind), **kw)

        if output_array is not None:
            output_array[...] = self.output_array
            return output_array
        else:
            return self.output_array


class TensorProductSpace(PFFT):
    """Class for multidimensional tensorproductspaces.

    The tensorproductspaces are created as Cartesian products from a set of 1D
    bases. The 1D bases are subclassed instances of the :class:`.SpectralBase`
    class.

    Parameters
    ----------
    comm : MPI communicator
    bases : list
        List of 1D bases (:class:`.SpectralBase`)
    axes : tuple of ints, optional
        A tuple containing the order of which to perform transforms.
        Last item is transformed first. Defaults to range(len(bases))
    dtype : data-type, optional
        Type of input data in real physical space. If not provided it
        will be inferred from the bases.
    slab : bool, optional
        Use 1D slab decomposition.
    collapse_fourier : bool, optional
        Collapse axes for Fourier bases if possible
    backward_from_pencil : False or Pencil
        In case of Pencil configure the transform by starting from the
        Pencil distribution in spectral space. This is primarily intended
        for a padded space, where the spectral distribution must be
        equal to the non-padded space.
    coordinates : 2- or 3-tuple, optional
        Map for curvilinear coordinatesystem.
        First tuple are the coordinate variables in the new coordinate system
        Second tuple are the Cartesian coordinates as functions of the
        variables in the first tuple. Example::

            psi = (theta, r) = sp.symbols('x,y', real=True, positive=True)
            rv = (r*sp.cos(theta), r*sp.sin(theta))

        where psi and rv are the first and second tuples, respectively.
        If a third item is provided with the tuple, then this third item
        is used as an additional assumption. For example, it is necessary
        to provide the assumption ``sympy.Q.positive(sympy.sin(theta))``,
        such that sympy can evaluate ``sqrt(sympy.sin(theta)**2)`` to
        ``sympy.sin(theta)`` and not ``Abs(sympy.sin(theta))``. Different
        coordinates may require different assumptions to help sympy when
        computing basis functions etc. See :class:`~shenfun.coordinates.Coordinates`.
    modify_spaces_inplace : bool, optional
        Whether or not a copy should be made of the input functionspaces.
        If True, then the input spaces will be modified inplace.
    kw : dict, optional
        Dictionary that can be used to plan transforms. Input to method
        ``plan`` for the bases.

    """
    def __init__(self, comm, bases, axes=None, dtype=None, slab=False,
                 collapse_fourier=False, backward_from_pencil=False,
                 coordinates=None, modify_spaces_inplace=False, **kw):
        # Note do not call __init__ of super
        self.comm = comm
        self.bases = bases
        if not modify_spaces_inplace:
            self.bases = tuple([base.get_unplanned() for base in bases])

        coors = coordinates
        if coors is None:
            psi = sp.symbols('x,y,z,r,s', real=True)
            coors = (psi[:len(bases)],)*2

        self.coors = Coordinates(*coors)
        self.hi = self.coors.hi
        self.sg = self.coors.sg
        shape = list(self.global_shape())
        self.axes = axes
        assert shape
        #assert min(shape) > 0
        if min(shape) == 0:
            self.subcomm = comm
            self._dtype = dtype
            return

        if axes is not None:
            axes = list(axes) if not isinstance(axes, int) else [axes]
        else:
            axes = list(range(len(shape)))

        for i, ax in enumerate(axes):
            if isinstance(ax, (int, np.integer)):
                if ax < 0:
                    ax += len(shape)
                axes[i] = (ax,)
            else:
                assert isinstance(ax, (tuple, list))
                ax = list(ax)
                for j, a in enumerate(ax):
                    assert isinstance(a, int)
                    if a < 0:
                        a += len(shape)
                        ax[j] = a
                axes[i] = ax
            assert min(axes[i]) >= 0
            assert max(axes[i]) < len(shape)
            assert 0 < len(axes[i]) <= len(shape)
            assert sorted(axes[i]) == sorted(set(axes[i]))

        if dtype is None:
            dtype = complex if isinstance(self.bases[axes[-1][-1]], C2C) else float

        dtype = np.dtype(dtype)
        assert dtype.char in 'fdgFDG'

        self.xfftn = []
        self.transfer = []
        self.pencil = [None, None]
        for axis, base in enumerate(self.bases):
            base.tensorproductspace = self
            base.axis = axis
            base.sl = slicedict(axis=axis, dimensions=len(self.bases))
            base.si = islicedict(axis=axis, dimensions=len(self.bases))

        self.axes = axes
        if not backward_from_pencil:
            if isinstance(self.bases[axes[-1][-1]], C2C):
                assert np.dtype(dtype).char in 'FDG'
            elif isinstance(self.bases[axes[-1][-1]], R2C):
                assert np.dtype(dtype).char in 'fdg'

            if isinstance(comm, tuple):
                assert slab is False
                assert len(comm) == len(shape)
                assert comm[axes[-1][-1]].Get_size() == 1
                self.subcomm = comm
            else:
                if slab:
                    axis = (axes[-1][-1] + 1) % len(shape)
                    dims = [1] * len(shape)
                    dims[axis] = comm.Get_size()
                else:
                    dims = [0] * len(shape)
                    for ax in axes[-1]:
                        dims[ax] = 1
                self.subcomm = Subcomm(comm, dims)

            # At this points axes is a tuple of tuples of length one.
            # Try to collapse some Fourier transforms into one.
            if np.any([abs(base.padding_factor - 1.0) > 1e-6 for base in self.bases]):
                collapse_fourier = False
            if collapse_fourier:
                F = lambda ax: self.bases[ax].family() == 'fourier' and self.subcomm[ax].Get_size() == 1
                axis = self.axes[-1][-1]
                groups = [list(self.axes[-1])]
                F0 = F(axis)
                for ax in reversed(self.axes[:-1]):
                    axis = ax[-1]
                    if F0 and F(axis):
                        groups[0].insert(0, axis)
                    else:
                        groups.insert(0, list(ax))
                    F0 = F(axis)
                self.axes = groups
            self.axes = tuple(map(tuple, self.axes))

            # Configure all transforms
            axes = self.axes[-1]
            pencil = Pencil(self.subcomm, shape, axes[-1])
            self.xfftn.append(self.bases[axes[-1]])
            self.xfftn[-1].plan(pencil.subshape, axes, dtype, kw)
            self.pencil[0] = pencilA = pencil
            if not (shape[axes[-1]] == self.xfftn[-1].forward.output_array.shape[axes[-1]] and
                    self.xfftn[-1].forward.input_array.dtype == self.xfftn[-1].forward.output_array.dtype):
                dtype = self.xfftn[-1].forward.output_array.dtype
                shape[axes[-1]] = self.xfftn[-1].forward.output_array.shape[axes[-1]]
                pencilA = Pencil(self.subcomm, shape, axes[-1])

            for i, axes in enumerate(reversed(self.axes[:-1])):
                pencilB = pencilA.pencil(axes[-1])
                transAB = pencilA.transfer(pencilB, dtype)
                xfftn = self.bases[axes[-1]]
                xfftn.plan(pencilB.subshape, axes, dtype, kw)
                self.xfftn.append(xfftn)
                self.transfer.append(transAB)
                pencilA = pencilB
                if not (shape[axes[-1]] == xfftn.forward.output_array.shape[axes[-1]] and
                        xfftn.forward.input_array.dtype == xfftn.forward.output_array.dtype):
                    dtype = xfftn.forward.output_array.dtype
                    shape[axes[-1]] = xfftn.forward.output_array.shape[axes[-1]]
                    pencilA = Pencil(pencilB.subcomm, shape, axes[-1])

            self.pencil[1] = pencilA

            self.forward = ForwardTransform(
                [o.forward for o in self.xfftn],
                [o.forward for o in self.transfer],
                self.pencil, self)
            self.backward = BackwardTransform(
                [o.backward for o in self.xfftn[::-1]],
                [o.backward for o in self.transfer[::-1]],
                self.pencil[::-1])
            self.scalar_product = ScalarTransform(
                [o.scalar_product for o in self.xfftn],
                [o.forward for o in self.transfer],
                self.pencil, self)

        else:
            self.configure_backwards(backward_from_pencil, dtype, kw)

        for i, base in enumerate(self.bases):
            base.axis = i
            if base.has_nonhomogeneous_bcs:
                base.bc.set_tensor_bcs(base, self)

    def configure_backwards(self, pencil, dtype, kw):
        """Configure transforms starting from spectral space

        Parameters
        ----------
        pencil : :class:`.Pencil`
            The distribution in spectral space
        dtype : Numpy dtype
            The type of data in spectral space
        kw : dict
            Any parameters for planning transforms

        Note
        ----
        To ensure the same distribution in spectral space, the padded
        space must be configured by moving from the spectral space towards the
        physical space. The distribution does not have to agree in physical
        space, because the padding is done only in the spectral.
        """
        shape = list(self.global_shape(True))
        axes = self.axes[0]
        xfftn = self.bases[axes[-1]]
        self.xfftn.append(xfftn)
        self.subcomm = pencil.subcomm
        subshape = list(pencil.subshape)
        if isinstance(xfftn, R2C):
            subshape[axes[-1]] = int(np.floor(xfftn.N*xfftn.padding_factor))
            dtype = np.dtype(dtype.char.lower())
        else:
            subshape[axes[-1]] = int(np.floor(subshape[axes[-1]]*xfftn.padding_factor))
        self.xfftn[-1].plan(subshape, axes, dtype, kw)
        if not (shape[axes[-1]] == self.xfftn[-1].forward.input_array.shape[axes[-1]] and
                self.xfftn[-1].forward.input_array.dtype == self.xfftn[-1].forward.output_array.dtype):
            dtype = self.xfftn[-1].forward.input_array.dtype
            shape[axes[-1]] = self.xfftn[-1].forward.input_array.shape[axes[-1]]
        pencilA = Pencil(pencil.subcomm, shape, axes[0])
        self.pencil[0] = pencilA
        for axes in self.axes[1:]:
            pencilB = pencilA.pencil(axes[-1])
            transBA = pencilA.transfer(pencilB, dtype)
            xfftn = self.bases[axes[-1]]
            subshape = list(pencilB.subshape)
            if isinstance(xfftn, R2C):
                subshape[axes[-1]] = int(np.floor(xfftn.N*xfftn.padding_factor))
                dtype = np.dtype(dtype.char.lower())
            else:
                subshape[axes[-1]] = int(np.floor(subshape[axes[-1]]*xfftn.padding_factor))
            xfftn.plan(subshape, axes, dtype, kw)
            self.xfftn.append(xfftn)
            self.transfer.append(transBA)
            pencilA = pencilB
            if not (shape[axes[-1]] == xfftn.forward.input_array.shape[axes[-1]] and
                    xfftn.forward.input_array.dtype == xfftn.forward.output_array.dtype):
                dtype = xfftn.forward.input_array.dtype
                shape[axes[-1]] = xfftn.forward.input_array.shape[axes[-1]]
                pencilA = Pencil(pencilB.subcomm, shape, axes[-1])

        self.pencil[1] = pencilA

        self.backward = BackwardTransform(
            [o.backward for o in self.xfftn],
            [o.forward for o in self.transfer],
            self.pencil)
        self.forward = ForwardTransform(
            [o.forward for o in self.xfftn[::-1]],
            [o.backward for o in self.transfer[::-1]],
            self.pencil[::-1], self)
        self.scalar_product = ScalarTransform(
            [o.scalar_product for o in self.xfftn[::-1]],
            [o.forward for o in self.transfer[::-1]],
            self.pencil[::-1], self)

    def get_dealiased(self, padding_factor=1.5, dealias_direct=False):
        """Return space (otherwise as self) to be used for dealiasing

        Parameters
        ----------
        padding_factor : number or tuple, optional
            Scale the number of quadrature points in self with this number, or
            tuple of numbers, one for each dimension.
        dealias_direct : bool, optional
            Used only if ``padding_factor=1``. Sets the 1/3 highest frequencies
            to zeros.
        """
        if padding_factor == 1 and dealias_direct is False:
            return self
        if isinstance(padding_factor, Number):
            padding_factor = (padding_factor,)*len(self)
        elif isinstance(padding_factor, (tuple, list, np.ndarray)):
            assert len(padding_factor) == len(self)
        padded_bases = [base.get_dealiased(padding_factor=padding_factor[axis],
                                           dealias_direct=dealias_direct)
                        for axis, base in enumerate(self.bases)]
        # Need the correct order of the transforms in case reversed somehow
        axes = []
        for ax in self.axes:
            for ai in ax:
                axes.append(ai)
        return TensorProductSpace(self.comm, padded_bases, axes=tuple(axes),
                                  dtype=self.forward.output_array.dtype,
                                  backward_from_pencil=self.forward.output_pencil,
                                  coordinates=self.coors.coordinates)

    def get_refined(self, N):
        """Return space (otherwise as self) refined to new shape

        Parameters
        ----------
        N : Number or sequence of integers
            The global shape of the new space. If N is a number, then
            refine the global shape by multiplying self's global shape by N.
            If N is an array, then N is treated as the new global shape of
            the returned space.
        """
        if isinstance(N, Number):
            N = N*np.array(self.global_shape())
        elif isinstance(N, (tuple, list, np.ndarray)):
            assert len(N) == len(self)
        refined_bases = [base.get_refined(N[axis])
                         for axis, base in enumerate(self.bases)]
        return TensorProductSpace(self.subcomm, refined_bases, axes=self.axes,
                                  dtype=self.dtype(),
                                  coordinates=self.coors.coordinates)

    def get_unplanned(self, tensorproductspace=False, **kwargs):
        """Return unplanned bases otherwise as self. Or return a new
        TensorProductSpace from the unplanned bases.

        Parameters
        ----------
        tensorproductspace : boolean
            if True, then return a new TensorProductSpace
            if False, return only the bases
        kwargs : keyword arguments
            Any arguments used in the creation of the different bases.

        """
        bases = tuple([base.get_unplanned(**kwargs) for base in self.bases])
        if tensorproductspace:
            return TensorProductSpace(comm, bases)
        return bases

    def get_homogeneous(self, **kwargs):
        """Return space with homogeneous boundary conditions, and otherwise
        as self.

        Parameters
        ----------
        kwargs : keyword arguments
            Any arguments used in the creation of the different bases.

        """
        bases = tuple([base.get_homogeneous(**kwargs) for base in self.bases])
        return TensorProductSpace(comm, bases)

    def dtype(self, forward_output=False):
        if hasattr(self, 'forward'):
            return PFFT.dtype(self, forward_output)
        return self._dtype

    def convolve(self, a_hat, b_hat, ab_hat):
        """Convolution of a_hat and b_hat

        Parameters
        ----------
        a_hat : array
            Input array of shape and type as output array from
            self.forward, or instance of :class:`.Function`
        b_hat : array
            Input array of shape and type as output array from
            self.forward, or instance of :class:`.Function`
        ab_hat : array
            Return array of same type and shape as a_hat and b_hat

        Note
        ----
        The return array ab_hat is truncated to the shape of a_hat and b_hat.

        Also note that self should have bases with padding for this method to give
        a convolution without aliasing. The padding is specified when creating
        instances of bases for the :class:`.TensorProductSpace`.

        FIXME Efficiency due to allocation
        """
        a = self.backward.output_array.copy()
        b = self.backward.output_array.copy()
        a = self.backward(a_hat, a)
        b = self.backward(b_hat, b)
        ab_hat = self.forward(a*b, ab_hat)
        return ab_hat

    def eval(self, points, coefficients, output_array=None, method=1):
        """Evaluate Function at points, given expansion coefficients

        Parameters
        ----------
        points : float or array of floats
            Array must be of shape (D, N), for N points in D dimensions
        coefficients : array
            Expansion coefficients, or instance of :class:`.Function`
        output_array : array, optional
            Return array, function values at points
        method : int, optional
            Chooses implementation. The method 0 is a low-memory cython
            version. Using method = 1 (default) leads to a faster cython
            implementation that, on the downside, uses more memory.
            The final, method = 2, is a python implementation.
        """
        if output_array is None:
            output_array = np.zeros(points.shape[1], dtype=self.forward.input_array.dtype)
        else:
            output_array[:] = 0
        if len(self.get_nonperiodic_axes()) > 1:
            method = 1
        assert self.dimensions < 4, 'eval not implemented (yet) for higher dimensions'
        if method == 0:
            return self._eval_lm_cython(points, coefficients, output_array)
        elif method == 1:
            return self._eval_cython(points, coefficients, output_array)
        else:
            return self._eval_python(points, coefficients, output_array)

    def _eval_python(self, points, coefficients, output_array):
        """Evaluate Function at points, given expansion coefficients

        Parameters
        ----------
        points : float or array of floats
        coefficients : array
            Expansion coefficients
        output_array : array
            Return array, function values at points
        """
        P = []
        last_conj_index = -1
        sl = -1
        out = []
        previous_axes = []
        flataxes = []
        for ax in self.axes:
            for a in ax:
                flataxes.append(a)
        for axis in flataxes:
            base = self.bases[axis]
            x = base.map_reference_domain(points[axis])
            D = base.evaluate_basis_all(x=x, argument=1)
            P = D[..., self.local_slice()[axis]]
            if isinstance(base, R2C):
                M = base.N//2+1
                if base.N % 2 == 0:
                    last_conj_index = M-1
                else:
                    last_conj_index = M
                offset = self.local_slice()[axis].start
                sl = offset
                st = self.local_slice()[axis].stop
                if sl == 0:
                    sl = 1
                st = min(last_conj_index, st) - offset
                sl -= offset
                sp = [slice(None), slice(sl, st)]

            if len(out) == 0:
                out = np.tensordot(P, coefficients, (1, axis))
                if isinstance(base, R2C):
                    ss = [slice(None)]*len(self)
                    ss[axis] = slice(sl, st)
                    out += np.conj(np.tensordot(P[tuple(sp)], coefficients[tuple(ss)], (1, axis)))
                    out = out.real

            else:
                k = np.count_nonzero([m < axis for m in previous_axes])
                if len(self) == 2:
                    if not isinstance(base, R2C):
                        out2 = np.sum(P*out, axis=-1)
                    else:
                        sp = tuple(sp)
                        out2 = np.sum(P.real*out.real - P.imag*out.imag, axis=-1)
                        out2 += np.sum(np.conj(P[sp].real*out[sp].real - P[sp].imag*out[sp].imag), axis=-1)

                elif len(self) == 3:
                    sx = [slice(None)]*out.ndim
                    sd = [slice(None)]*out.ndim
                    if len(out.shape) == 3:
                        kk = 1 if axis-k == 1 else 2
                        sx[kk] = np.newaxis
                    if not isinstance(base, R2C):
                        out2 = np.sum(P[tuple(sx)]*out, axis=1+axis-k)
                    else:
                        out2 = np.sum(P[tuple(sx)].real*out.real - P[tuple(sx)].imag*out.imag, axis=1+axis-k)
                        sx[1+axis-k] = slice(sl, st)
                        sd[1+axis-k] = slice(sl, st)
                        sx = tuple(sx)
                        sd = tuple(sd)
                        out2 += np.sum(np.conj(P[sx].real*out[sd].real - P[sx].imag*out[sd].imag), axis=1+axis-k)
                out = out2
            previous_axes.append(axis)
        output_array[:] = out
        output_array = comm.allreduce(output_array)
        return output_array


    def _eval_lm_cython(self, points, coefficients, output_array):
        """Evaluate Function at points, given expansion coefficients

        Parameters
        ----------
        points : float or array of floats
        coefficients : array
            Expansion coefficients
        output_array : array
            Return array, function values at points
        """
        r2c = -1
        last_conj_index = -1
        sl = -1
        x = []
        w = []
        for base in self:
            axis = base.axis
            if isinstance(base, R2C):
                r2c = axis
                M = base.N//2+1
                if base.N % 2 == 0:
                    last_conj_index = M-1
                else:
                    last_conj_index = M
                sl = self.local_slice()[axis].start
            x.append(base.map_reference_domain(points[axis]))
            w.append(base.wavenumbers(bcast=False)[self.local_slice()[axis]].astype(float))

        if len(self) == 2:
            output_array = evaluate.evaluate_lm_2D(list(self.bases), output_array, coefficients, x[0], x[1], w[0], w[1], r2c, last_conj_index, sl)

        elif len(self) == 3:
            output_array = evaluate.evaluate_lm_3D(list(self.bases), output_array, coefficients, x[0], x[1], x[2], w[0], w[1], w[2], r2c, last_conj_index, sl)

        output_array = np.atleast_1d(output_array)
        output_array = comm.allreduce(output_array)
        return output_array

    def _eval_cython(self, points, coefficients, output_array):
        """Evaluate Function at points, given expansion coefficients

        Parameters
        ----------
        points : float or array of floats
        coefficients : array
            Expansion coefficients
        output_array : array
            Return array, function values at points
        """
        P = []
        r2c = -1
        last_conj_index = -1
        sl = -1
        for base in self:
            axis = base.axis
            x = base.map_reference_domain(points[axis])
            D = base.evaluate_basis_all(x=x, argument=1)
            P.append(D[..., self.local_slice()[axis]])
            if isinstance(base, R2C):
                r2c = axis
                M = base.N//2+1
                if base.N % 2 == 0:
                    last_conj_index = M-1
                else:
                    last_conj_index = M
                sl = self.local_slice()[axis].start
        if len(self) == 2:
            output_array = evaluate.evaluate_2D(output_array, coefficients, P, r2c, last_conj_index, sl)

        elif len(self) == 3:
            output_array = evaluate.evaluate_3D(output_array, coefficients, P, r2c, last_conj_index, sl)

        output_array = np.atleast_1d(output_array)
        output_array = comm.allreduce(output_array)
        return output_array

    def wavenumbers(self, scaled=False, eliminate_highest_freq=False):
        """Return list of wavenumbers

        Parameters
        ----------
        scaled : bool, optional
            Scale wavenumbers with size of box
        eliminate_highest_freq : bool, optional
            Set Nyquist frequency to zero for evenly shaped axes of Fourier
            bases.

        """
        K = []
        for base in self:
            K.append(base.wavenumbers(scaled=scaled,
                                      eliminate_highest_freq=eliminate_highest_freq))
        return K

    def local_wavenumbers(self, broadcast=False, scaled=False,
                          eliminate_highest_freq=False):
        """Return list of local wavenumbers

        Parameters
        ----------
        broadcast : bool, optional
            Broadcast returned wavenumber arrays to actual
            dimensions of :class:`.TensorProductSpace`
        scaled : bool, optional
            Scale wavenumbers with size of box
        eliminate_highest_freq : bool, optional
            Set Nyquist frequency to zero for evenly shaped axes of Fourier
            bases

        """
        k = self.wavenumbers(scaled=scaled, eliminate_highest_freq=eliminate_highest_freq)
        lk = []
        for axis, (n, s) in enumerate(zip(k, self.local_slice(True))):
            ss = [slice(None)]*len(k)
            ss[axis] = s
            lk.append(n[tuple(ss)])
        if broadcast is True:
            return [np.broadcast_to(m, self.shape()) for m in lk]
        return lk

    def mesh(self, map_true_domain=True, kind='quadrature'):
        """Return list of 1D physical meshes for each dimension of
        :class:`.TensorProductSpace`

        Parameters
        ----------
        map_true_domain : bool, optional
            Whether or not to map points to true domain
        kind : str, optional

            - 'quadrature' - Use quadrature mesh
            - 'uniform' - Use uniform mesh
        """
        X = []
        for base in self:
            X.append(base.mesh(bcast=True, map_true_domain=map_true_domain, kind=kind))
        return X

    def local_mesh(self, bcast=False, map_true_domain=True, kind='quadrature'):
        """Return list of local 1D physical meshes for each dimension of
        :class:`.TensorProductSpace`

        Parameters
        ----------
        bcast : bool, optional
            Broadcast each 1D mesh to real shape of
            :class:`.TensorProductSpace`
        map_true_domain : bool, optional
            Whether or not to map points to true domain
        kind : str, optional

            - 'quadrature' - Use quadrature mesh
            - 'uniform' - Use uniform mesh
        """
        mesh = self.mesh(map_true_domain=map_true_domain, kind=kind)
        lm = []
        for axis, (n, s) in enumerate(zip(mesh, self.local_slice(False))):
            ss = [slice(None)]*len(mesh)
            ss[axis] = s
            lm.append(n[tuple(ss)])
        if bcast is True:
            return [np.broadcast_to(m, self.shape(False)) for m in lm]
        return lm

    def local_cartesian_mesh(self, map_true_domain=True, kind='quadrature'):
        """Return local Cartesian mesh

        Parameters
        ----------
        map_true_domain : bool, optional
            Whether or not to map points to true domain
        kind : str, optional

            - 'quadrature' - Use quadrature mesh
            - 'uniform' - Use uniform mesh
        """
        X = self.local_mesh(bcast=True, map_true_domain=map_true_domain, kind=kind)
        xx = []
        psi = self.coors.coordinates[0]
        for rv in self.coors.coordinates[1]:
            xx.append(sp.lambdify(psi, rv)(*X))
        return xx

    def cartesian_mesh(self, map_true_domain=True, kind='quadrature'):
        """Return Cartesian mesh

        Parameters
        ----------
        map_true_domain : bool, optional
            Whether or not to map points to true domain
        kind : str, optional

            - 'quadrature' - Use quadrature mesh
            - 'uniform' - Use uniform mesh
        """
        X = self.mesh(map_true_domain=map_true_domain, kind=kind)
        xx = []
        psi = self.coors.coordinates[0]
        for rv in self.coors.coordinates[1]:
            xx.append(sp.lambdify(psi, rv)(*X))
        return xx

    def dim(self):
        """Return dimension of ``self`` (degrees of freedom)"""
        return np.prod(self.dims())

    def dims(self):
        """Return dimensions (degrees of freedom) of all bases in ``self``"""
        return tuple([base.dim() for base in self])

    def size(self, forward_output=False):
        """Return number of elements in :class:`.TensorProductSpace`"""
        return np.prod(self.shape(forward_output))

    def slice(self):
        """The slices of dofs for each dimension"""
        return tuple(base.slice() for base in self.bases)

    @property
    def use_fixed_gauge(self):
        """Return True if :class:`.TensorProductSpace` contains only spaces
        with Neumann boundary conditions.
        """
        for base in self.bases:
            if base.boundary_condition() != 'Neumann':
                # Any other boundary condition triggers False
                # It is only Neumann bases that requires a gauge
                return False
        return True

    def global_shape(self, forward_output=False):
        """Return global shape of arrays for :class:`.TensorProductSpace`

        Parameters
        ----------
        forward_output : bool, optional
            If True then return shape of an array that is the result of a
            forward transform. If False then return shape of physical
            space, i.e., the input to a forward transform.

        """
        return tuple([base.shape(forward_output) for base in self])

    @property
    def domains(self):
        """Return domains for all directions
        """
        return tuple([base.domain for base in self])

    @property
    def volume(self):
        """Return computational volume
        """
        vol = 1
        for d in self.domains:
            vol *= (d[1]-d[0])
        return vol

    def mask_nyquist(self, u_hat, mask=None):
        """Return :class:`.Function` ``u_hat`` with zero Nyquist coefficients

        Parameters
        ----------
        u_hat : array
            :class:`.Function` to be masked
        mask : array or None, optional
            mask array, if not provided then get the mask by calling
            :func:`get_mask_nyquist`
        """
        if mask is None:
            mask = self.get_mask_nyquist()
        u_hat = apply_mask(u_hat, mask)
        return u_hat

    def get_mask_nyquist(self):
        """Return an array with zeros for Nyquist coefficients and one otherwise"""
        k = []
        do_mask = False
        for base in self:
            if base.family() == 'fourier':
                k.append(base.get_mask_nyquist(bcast=True))
                if base.N % 2 == 0:
                    do_mask = True
            else:
                k.append(None)

        if not do_mask:
            return None

        lk = []
        for axis, (n, s) in enumerate(zip(k, self.local_slice(True))):
            if n is None:
                lk.append(1)
            else:
                ss = [slice(None)]*len(k)
                ss[axis] = s
                lk.append(n[tuple(ss)])

        mask = 1
        for ll in lk:
            mask = mask * ll
        return mask

    def get_measured_array(self, u):
        """Weigh Array ``u`` with integral measure

        Parameters
        ----------
        u : Array
        """
        dx = self.coors.get_sqrt_det_g()
        mesh = self.local_mesh(True)
        if dx == 1:
            return u

        if isinstance(dx, Number):
            u *= dx
            return u

        sym0 = dx.free_symbols
        m = []
        for sym in sym0:
            j = 'xyzrs'.index(str(sym))
            m.append(mesh[j])
        xj = sp.lambdify(tuple(sym0), dx)(*m)
        u *= xj
        return u

    def __iter__(self):
        return iter(self.bases)

    @property
    def rank(self):
        """Return tensor rank of :class:`.TensorProductSpace`"""
        return 0

    @property
    def tensor_rank(self):
        return 0

    def __len__(self):
        """Return dimension of :class:`.TensorProductSpace`"""
        return len(self.bases)

    def num_components(self):
        """Return number of scalar spaces in :class:`.TensorProductSpace`"""
        return 1

    @property
    def dimensions(self):
        """Return dimension of :class:`.TensorProductSpace`"""
        return self.__len__()

    def get_nonperiodic_axes(self):
        axes = []
        for axis, base in enumerate(self):
            if not base.family() == 'fourier':
                axes.append(axis)
        return np.array(axes)

    def get_nondiagonal_axes(self):
        # Return list of axes that **may** contain non-diagonal matrices
        axes = []
        if isinstance(self.coors.hi.prod(), sp.Expr):
            x = 'xyzrs'
            sym = self.coors.hi.prod().free_symbols
            msaxes = set()
            for s in sym:
                if str(s) in x:
                    msaxes.add(x.index(str(s)))

        for axis, base in enumerate(self):
            if not base.family() == 'fourier':
                axes.append(axis)
            if base.family() == 'fourier' and (axis in msaxes) and base.N > 1:
                axes.append(axis)
        return np.array(axes)

    def get_diagonal_axes(self):
        return np.setxor1d(self.get_nondiagonal_axes(), range(self.dimensions)).astype(int)

    def get_nonhomogeneous_axes(self):
        axes = []
        for axis, base in enumerate(self):
            if base.has_nonhomogeneous_bcs:
                axes.append(axis)
        return np.array(axes)

    @property
    def is_orthogonal(self):
        ortho = True
        for base in self.bases:
            ortho *= base.is_orthogonal
        return ortho

    @property
    def is_padded(self):
        padded = True
        for base in self.bases:
            padded *= base.is_padded
        return padded

    def get_orthogonal(self):
        """Return tensor product space using the orthogonal basis for all directions"""
        ortho = []
        for base in self.bases:
            ortho.append(base.get_orthogonal())
        return TensorProductSpace(self.subcomm, ortho, axes=self.axes,
                                  dtype=self.forward.input_array.dtype,
                                  coordinates=self.coors.coordinates)

    def get_testspace(self, kind='Galerkin', **kwargs):
        r"""Return appropriate test space

        Parameters
        ----------
        kind : str, optional

            - 'Galerkin' - or 'G'
            - 'Petrov-Galerkin' - or 'PG'

            If kind = 'Galerkin', then return self, corresponding to a regular
            Galerkin method.

            If kind = 'Petrov-Galerkin' or 'PG', then return a test space that makes
            use of the testbases `Phi1`, `Phi2`, `Phi3` or `Phi4`, where the choice
            is made based on the number of boundary conditions in self. One boundary
            condition uses `Phi1`, two uses `Phi2` etc.

        kwargs : keyword arguments, optional
            Any other keyword arguments used in the creation of the test bases.

        Note
        ----
        Petrov-Galerkin is only possible for Jacobi polynomials. If mixed with,
        e.g., Fourier spaces, then regular Galerkin is used for the non-polynomial
        bases.

        Returns
        -------
        Instance of :class:`.TensorProductSpace`
        """
        test = []
        for base in self.bases:
            test.append(base.get_testspace(kind=kind, **kwargs))
        return TensorProductSpace(self.subcomm, test, axes=self.axes,
                                  dtype=self.forward.input_array.dtype,
                                  coordinates=self.coors.coordinates)

    def get_adaptive(self, fun=None, reltol=1e-12, abstol=1e-15):
        """Return space (otherwise as self) with number of quadrature points
        determined by fitting `fun`

        Returns
        -------
        :class:`.TensorProductSpace`
            A new space with adaptively found number of quadrature points

        Note
        ----
        Only spaces defined with N=0 quadrature points are adapted, the
        others are kept as is

        Example
        -------
        >>> from shenfun import FunctionSpace, Function, TensorProductSpace, comm
        >>> import sympy as sp
        >>> r, theta = psi = sp.symbols('x,y', real=True, positive=True)
        >>> rv = (r*sp.cos(theta), r*sp.sin(theta))
        >>> B0 = FunctionSpace(0, 'C', domain=(0, 1))
        >>> F0 = FunctionSpace(0, 'F', dtype='d')
        >>> T = TensorProductSpace(comm, (B0, F0), coordinates=(psi, rv))
        >>> u = Function(T, buffer=((1-r)*r)**4*(sp.cos(10*theta)))
        >>> print(u.global_shape)
        (9, 11)

        """
        domains = [base.domain for base in self.bases]
        sym0 = fun.free_symbols
        sd = {str(sym): sym for sym in sym0}
        Tj = []
        for axis, base in enumerate(self.bases):
            # First remove all other axes from the function
            otheraxes = list(range(len(self.bases)))
            otheraxes.remove(axis)
            fj = fun.copy()
            bcc = None
            if base.has_nonhomogeneous_bcs:
                bcc = base.bc.bcs.copy()
            for ax in otheraxes:
                fj = fj.subs(sd['xyzrs'[ax]], 0.5*(domains[ax][1]-domains[ax][0]))
            if base.N == 0:
                par = dict(fun=fj, reltol=reltol, abstol=abstol)
                if base.has_nonhomogeneous_bcs:
                    for l, bc in enumerate(bcc):
                        bc = bc.subs(sd['xyzrs'[ax]], 0.5*(domains[ax][1]-domains[ax][0]))
                        bcc[l] = bc
                    par['bc'] = bcc
                Tj.append(base.get_adaptive(**par))
                if base.has_nonhomogeneous_bcs:
                    Tj[-1].bc = BoundaryValues(Tj[-1], bc=base.bc)
            else:
                Tj.append(base)
        return TensorProductSpace(self.subcomm, Tj, axes=self.axes,
                                  coordinates=self.coors.coordinates)

    @property
    def is_composite_space(self):
        return 0

    def __getitem__(self, i):
        """Return instance of base i

        Parameters
        ----------
        i : int
        """
        return self.bases[i]

    def compatible_base(self, space):
        """Return whether space is compatible with self.

        Parameters
        ----------

        space : :class:`.TensorProductSpace`
            The space compared to

        Note
        ----
        Two spaces are deemed compatible if the underlying bases, along each
        direction, belong to the same family, has the same number of quadrature
        points, and the same quadrature scheme.
        """
        compatible = True
        for base0, base1 in zip(self.bases, space.bases):
            if not hash(base0) == hash(base1):
                compatible = False
                break
        return compatible

    def shape(self, forward_output=True):
        """The local (to each processor) shape of data

        Parameters
        ----------
        forward_output : bool, optional
            Return shape of output array (spectral space) if True, else return
            shape of input array (physical space)
        """
        if forward_output is not True:
            return self.forward.input_pencil.subshape
        return self.forward.output_array.shape

    def _get_ndiag_cum_dofs(self):
        """Return the cumulative sum of degrees of freedom along nondiagonal axes"""
        dims = np.array(self.dims())
        daxes = self.get_diagonal_axes()
        if len(daxes) > 0:
            dims[daxes] = 1
        return np.array([0, np.prod(dims)], dtype=int)

    def _get_ndiag_slices(self, j=()):
        """Return a sequence of slices for non-diagonal axes and integers for diagonal"""
        assert self.num_components() == 1
        sl = (0,)+self.slice()
        daxes = self.get_diagonal_axes()
        if len(daxes) > 0:
            daxes += 1
        j = np.atleast_1d((0,)*len(daxes) if j == () else j)
        assert len(j) == len(daxes)
        sa = np.array(sl)
        if len(daxes) > 0:
            sa.T[daxes] = j if isinstance(j, int) else np.array(j)[:, None]
        return np.atleast_2d(sa)

    def _get_ndiag_slices_and_dims(self, j=()):
        """Return nondiagonal slices and their accumulated dimensions"""
        sl = self._get_ndiag_slices(j)
        dims = self._get_ndiag_cum_dofs()
        return sl, dims


class CompositeSpace:
    """Class for composite tensorproductspaces.

    The mixed spaces are Cartesian products of :class:`.TensorProductSpace` or
    other instances of :class:`.CompositeSpace`.

    Parameters
    ----------
    spaces : list
        List of spaces
    """

    def __init__(self, spaces):
        self.spaces = spaces
        self.forward = VectorTransform([space.forward for space in spaces])
        self.backward = VectorTransform([space.backward for space in spaces])
        self.scalar_product = VectorTransform([space.scalar_product for space in spaces])

    @property
    def is_composite_space(self):
        return 1

    def eval(self, points, coefficients, output_array=None, method=1):
        """Evaluate Function at points, given expansion coefficients

        Parameters
        ----------
        points : float or array of floats
        coefficients : array
            Expansion coefficients
        output_array : array, optional
            Return array, function values at points
        method : int, optional
            Chooses implementation. Method 0 is a low-memory cython
            version. Using method = 1 (default) leads to a faster cython
            implementation that, on the downside, uses more memory.
            The final, method = 2, is a python implementation used mainly
            for verification.
        """
        if output_array is None:
            output_array = np.zeros((len(self.flatten()), points.shape[-1]), dtype=self.forward.input_array.dtype)
        for i, space in enumerate(self.flatten()):
            output_array.__array__()[i] = space.eval(points, coefficients.__array__()[i], output_array.__array__()[i], method)
        return output_array

    def convolve(self, a_hat, b_hat, ab_hat):
        """Convolution of a_hat and b_hat

        Parameters
        ----------
        a_hat : array
            Input array of shape and type as output array from
            self.forward, or instance of :class:`.Function`
        b_hat : array
            Input array of shape and type as output array from
            self.forward, or instance of :class:`.Function`
        ab_hat : array
            Return array of same type and shape as a_hat and b_hat

        Note
        ----
        Note that self should have bases with padding for this method to give
        a convolution without aliasing. The padding is specified when creating
        instances of bases for the TensorProductSpace.

        FIXME Efficiency due to allocation
        """
        N = list(self.backward.output_array.shape)
        a = np.zeros([self.dimensions]+N, dtype=self.backward.output_array.dtype)
        b = np.zeros([self.dimensions]+N, dtype=self.backward.output_array.dtype)
        a = self.backward(a_hat, a)
        b = self.backward(b_hat, b)
        ab_hat = self.forward(a*b, ab_hat)
        return ab_hat

    @property
    def dimensions(self):
        return self.flatten()[0].dimensions

    @property
    def rank(self):
        """Return rank of space

        Note
        ----
        This is 1 for all composite spaces, since we use a flat
        storage of all composite arrays.
        """
        return 1

    @property
    def tensor_rank(self):
        """Return rank of tensor

        Note
        ----
        This is the number of free indices in the tensor, 0 for scalar,
        1 for vector etc. It is None for a composite space that is not a
        tensor.
        """
        return None

    def dim(self):
        """Return dimension of ``self`` (degrees of freedom)"""
        return sum(self.dims())

    def dims(self):
        """Return dimensions (degrees of freedom) of all spaces in ``self``"""
        s = []
        for space in self.flatten():
            s.append(space.dim())
        return s

    def dims_composite(self):
        """Return dimensions per axis (degrees of freedom) of all spaces in ``self``"""
        s = []
        for space in self.flatten():
            s.append(space.dims())
        return np.array(s)

    def get_offsets(self):
        """Return offsets for spaces"""
        return np.concatenate((np.zeros((1, self.dimensions), dtype=int),
                               np.cumsum(self.dims_composite(), axis=0)))

    def _get_ndiag_cum_dofs(self):
        """Return the cumulative sum of degrees of freedom along nondiagonal axes"""
        dc = self.dims_composite()
        daxes = self.get_diagonal_axes()
        dc.T[daxes] = 1
        dci = np.prod(dc, axis=1)
        dims = np.array([0]+np.cumsum(dci).tolist())
        return dims

    def _get_ndiag_slices(self, j=()):
        sl = self.slice()
        daxes = self.get_diagonal_axes()
        if len(daxes) > 0:
            daxes += 1
        j = np.atleast_1d((0,)*len(daxes) if j == () else j)
        assert len(j) == len(daxes)
        sa = np.array(sl)
        if len(daxes) > 0:
            sa.T[daxes] = np.array(j)[:, None]
        return sa

    def _get_ndiag_slices_and_dims(self, j=()):
        sl = self._get_ndiag_slices(j)
        dims = [0]
        for s in sl:
            d = np.prod([1 if isinstance(i, int) else i.stop-i.start for i in s])
            dims.append(dims[-1]+d)
        return sl, np.array(dims)

    def slice(self):
        """The slices of dofs for each dimension"""
        return tuple(tuple([i]+list(space.slice())) for i, space in enumerate(self.flatten()))

    def size(self, forward_output=False):
        """Return number of elements in :class:`.CompositeSpace`"""
        N = self.shape(forward_output)
        if forward_output:
            return np.sum([np.prod(s) for s in N])
        else:
            return np.prod(N)

    def shape(self, forward_output=False):
        """Return shape of arrays for :class:`.CompositeSpace`

        Parameters
        ----------
        forward_output : bool, optional
            If True then return shape of an array that is the result of a
            forward transform. If False then return shape of physical
            space, i.e., the input to a forward transform.

        Note
        ----
        A :class:`.CompositeSpace` may contain tensor product spaces
        of different shape in spectral space. Hence this function returns a
        list of shapes and not one single tuple.
        """
        if forward_output:
            s = []
            for space in self.flatten():
                s.append(space.shape(forward_output))
        else:
            s = self.flatten()[0].shape(forward_output)
            s = (self.num_components(),) + s
        return s

    def global_shape(self, forward_output=False):
        """Return global shape for :class:`.CompositeSpace`

        Parameters
        ----------
        forward_output : bool, optional
            If True then return shape of an array that is the result of a
            forward transform. If False then return shape of physical
            space, i.e., the input to a forward transform.

        """
        s = self.flatten()[0].global_shape(forward_output)
        return (self.num_components(),) + s

    def local_slice(self, forward_output=False):
        """The local view into the global data

        Parameters
        ----------
        forward_output : bool, optional
            Return local slices of output array (spectral space) if True, else
            return local slices of input array (physical space)

        """
        s = self.flatten()[0].local_slice(forward_output)
        return (slice(None),) + s

    def num_components(self):
        f = self.flatten()
        return len(f)

    def flatten(self):
        s = []
        def _recursiveflatten(l, s):
            if hasattr(l, 'spaces'):
                for i in l.spaces:
                    _recursiveflatten(i, s)
            else:
                s.append(l)
        _recursiveflatten(self, s)
        return s

    def compatible_base(self, space):
        assert self.rank == space.rank
        compatible = True
        for space0, space1 in zip(self.flatten(), space.flatten()):
            if space0.compatible_base(space1) is False:
                compatible = False
                break
        return compatible

    def get_refined(self, N):
        raise NotImplementedError

    def get_orthogonal(self):
        raise NotImplementedError

    def get_dealiased(self, padding_factor=1.5, dealias_direct=False):
        raise NotImplementedError

    def __getitem__(self, i):
        return self.spaces[i]

    def __getattr__(self, name):
        obj = object.__getattribute__(self, 'spaces')
        assert name not in ('bases',)
        return getattr(obj[0], name)

    def __len__(self):
        return len(self.spaces)


class VectorSpace(CompositeSpace):
    """Vector of :class:`.TensorProductSpace` instances.

    This is simply a special :class:`.CompositeSpace` where the number of
    instances of the :class:`.TensorProductSpace` class equals the geometrical
    dimension of the problem.

    For example, a TensorProductSpace created by a tensorproduct of 2 1D
    spaces, will have vectors of length 2. A TensorProductSpace created from 3
    1D spaces will have vectors of length 3.

    Parameters
    ----------
    space : :class:`.TensorProductSpace` or list of ndim instances of the
    :class:`.TensorProductSpace` class to create vector from.

    """

    def __init__(self, space):
        if isinstance(space, list):
            assert len(space) == space[0].dimensions
            spaces = space
        else:
            spaces = [space]*space.dimensions
        CompositeSpace.__init__(self, spaces)

    def num_components(self):
        """Return number of spaces in mixed space"""
        assert len(self.spaces) == self.dimensions
        return self.dimensions

    @property
    def tensor_rank(self):
        return 1

    def shape(self, forward_output=False):
        """Return shape of arrays for :class:`.VectorSpace`

        Parameters
        ----------
        forward_output : bool, optional
            If True then return shape of an array that is the result of a
            forward transform. If False then return shape of physical
            space, i.e., the input to a forward transform.

        """
        s = self.flatten()[0].shape(forward_output)
        s = (self.num_components(),) + s
        return s

    def get_refined(self, N):
        if np.all([s == self.spaces[0] for s in self.spaces[1:]]):
            return VectorSpace(self.spaces[0].get_refined(N))
        return VectorSpace([s.get_refined(N) for s in self.spaces])

    def get_dealiased(self, padding_factor=1.5, dealias_direct=False):
        if padding_factor == 1 and dealias_direct == False:
            return self
        if np.all([s == self.spaces[0] for s in self.spaces[1:]]):
            return VectorSpace(self.spaces[0].get_dealiased(padding_factor, dealias_direct))
        return VectorSpace([s.get_dealiased(padding_factor, dealias_direct) for s in self.spaces])

    def get_orthogonal(self):
        if np.all([s == self.spaces[0] for s in self.spaces[1:]]):
            return VectorSpace(self.spaces[0].get_orthogonal())
        return VectorSpace([s.get_orthogonal() for s in self.spaces])

class TensorSpace(VectorSpace):
    """A special :class:`.CompositeSpace` for second rank tensors.

    A TensorProductSpace created by a tensorproduct of 2 1D
    bases, will have tensors of length 4. A TensorProductSpace created from 3
    1D bases will have tensors of length 9.

    Parameters
    ----------
    space : :class:`.TensorProductSpace` or list of ndim :class:`.VectorSpace`s
        Spaces to create vector from

    """

    def __init__(self, space):
        if isinstance(space, list):
            assert isinstance(space[0], VectorSpace)
            assert len(space) == space[0].dimensions
            spaces = space
        elif isinstance(space, VectorSpace):
            spaces = [space]*space.dimensions
        else:
            assert isinstance(space, TensorProductSpace)
            spaces = [VectorSpace(space)]*space.dimensions
        CompositeSpace.__init__(self, spaces)

    def num_components(self):
        return self.dimensions**2

    @property
    def tensor_rank(self):
        return 2

    def get_refined(self, N):
        if np.all([s == self.spaces[0] for s in self.spaces[1:]]):
            return TensorSpace(self.spaces[0].get_refined(N))
        return TensorSpace([s.get_refined(N) for s in self.spaces])

    def get_dealiased(self, padding_factor=1.5, dealias_direct=False):
        if padding_factor == 1 and dealias_direct == False:
            return self
        if np.all([s == self.spaces[0] for s in self.spaces[1:]]):
            return TensorSpace(self.spaces[0].get_dealiased(padding_factor, dealias_direct))
        return TensorSpace([s.get_dealiased(padding_factor, dealias_direct) for s in self.spaces])

    def get_orthogonal(self):
        if np.all([s == self.spaces[0] for s in self.spaces[1:]]):
            return TensorSpace(self.spaces[0].get_orthogonal())
        return TensorSpace([s.get_orthogonal() for s in self.spaces])

class VectorTransform:

    __slots__ = ('_transforms',)

    def __init__(self, transforms):
        self._transforms = []
        for transform in transforms:
            if isinstance(transform, VectorTransform):
                self._transforms += transform._transforms
            else:
                self._transforms.append(transform)

    def __getattr__(self, name):
        obj = object.__getattribute__(self, '_transforms')
        if name == '_transforms':
            return obj
        return getattr(obj[0], name)

    def __call__(self, input_array, output_array, kind=None, **kw):
        mesh = kw.get('mesh', None) # only backward transform
        for i, transform in enumerate(self._transforms):
            if mesh is not None:
                mi = mesh[i] if isinstance(mesh, CompositeSpace) else mesh
                kw['mesh'] = mi
            output_array.__array__()[i] = transform(input_array.__array__()[i],
                                                    output_array.__array__()[i],
                                                    kind=kind, **kw)
        return output_array


class Convolve:
    r"""Class for convolving without truncation.

    The convolution of :math:`\hat{a}` and :math:`\hat{b}` is computed by first
    transforming backwards with padding::

        a = Tp.backward(a_hat)
        b = Tp.backward(b_hat)

    and then transforming the product ``a*b`` forward without truncation::

        ab_hat = T.forward(a*b)

    where Tp is a :class:`.TensorProductSpace` for regular padding, and
    T is a TensorProductSpace with no padding, but using the shape
    of the padded a and b arrays.

    For convolve with truncation forward, use just the convolve method
    of the Tp space instead.

    Parameters
    ----------
    padding_space : :class:`.TensorProductSpace`
        Space with regular padding backward and truncation forward.
    """

    def __init__(self, padding_space):
        self.padding_space = padding_space
        shape = padding_space.shape()
        bases = []
        for i, base in enumerate(padding_space.bases):
            newbase = base.__class__(shape[i], padding_factor=1.0)
            bases.append(newbase)
        axes = []
        for axis in padding_space.axes:
            axes.append(axis[0])
        newspace = TensorProductSpace(padding_space.comm, bases, axes=axes)
        self.newspace = newspace

    def __call__(self, a_hat, b_hat, ab_hat=None):
        """Compute convolution of a_hat and b_hat without truncation

        Parameters
        ----------
        a_hat : :class:`.Function`
        b_hat : :class:`.Function`
        ab_hat : :class:`.Function`
        """
        Tp = self.padding_space
        T = self.newspace
        if ab_hat is None:
            ab_hat = Function(T)

        a = Array(Tp)
        b = Array(Tp)
        a = Tp.backward(a_hat, a)
        b = Tp.backward(b_hat, b)
        ab_hat = T.forward(a*b, ab_hat)
        return ab_hat


class BoundaryValues:
    """Class for setting nonhomogeneous boundary conditions inside a multi-
    dimensional :class:`.TensorProductSpace`.

    Parameters
    ----------
    T : :class:`.TensorProductSpace`
    bc : :class:`.BoundaryConditions`
    """
    # pylint: disable=protected-access, redefined-outer-name, dangerous-default-value, unsubscriptable-object

    def __init__(self, T, bc):
        self.base = T
        self.tensorproductspace = None
        self.bc = bc
        self.bcs = bc.orderedvals()       # Processed bc
        self.bcs_final = bc.orderedvals() # Data. May differ from bcs only for TensorProductSpaces
        self.axis = 0
        self.bc_time = 0

    def update(self, time=None):
        from shenfun.forms.project import Project
        tt = sp.symbols('t', real=True)
        bcs = self.bc.orderedvals()
        update_tensor = False
        if self.has_nonhomogeneous_bcs:
            for i, bci in enumerate(bcs):
                if isinstance(bci, Project):
                    bci()
                    update_tensor = True
                elif isinstance(bci, sp.Expr) and time is not None:
                    if tt in bci.free_symbols:
                        self.bc_time = time
                        bcs[i] = bci.subs(tt, time)
                        update_tensor = True
        if update_tensor:
            self.bcs = bcs
            self.bcs_final[:] = self.bcs
            if self.tensorproductspace is not None:
                self.set_tensor_bcs(self.base, self.tensorproductspace)

    def set_tensor_bcs(self, this_base, T):
        """Set correct boundary values for tensor, using values in self.bc

        Parameters
        ----------
        this_base : instance of :class:`.SpectralBase`
        T : instance of :class:`.TensorProductSpace`

        Note
        ----
        To modify boundary conditions on the fly, modify first self.bc and then
        call this function. The boundary condition can then be applied as before.
        """
        from shenfun import project
        from shenfun.utilities import split

        self.axis = this_base.axis
        self.base = this_base
        self.tensorproductspace = T

        if isinstance(T, SpectralBase):
            pass
        elif this_base.has_nonhomogeneous_bcs:
            # Setting an inhomogeneous boundary condition in a TensorProductSpace
            # is more involved than for a single dimension, and the routine will
            # depend on the order of the bases. Basically, the inhomogeneous value
            # needs to be projected to a space consisting of all the other function
            # spaces, except the inhomogeneous one. This would be very easy if it was
            # not for MPI.
            axis = None
            number_of_bases_after_this = 0
            for axes in reversed(T.axes):
                base = T.bases[axes[-1]]
                if base == this_base:
                    axis = base.axis
                else:
                    if axis is None:
                        number_of_bases_after_this += 1

            if self.has_nonhomogeneous_bcs() is False:
                for i in range(len(self.bcs)):
                    self.bcs[i] = self.bcs_final[i] = 0
                return

            # Set boundary values
            # These are values set at the end of a transform in an
            # inhomogeneous space, but before any other transforms

            b = Array(T)
            s = T.local_slice(False)[self.axis]
            num_bcs = self.bc.num_bcs()

            for j, bci in enumerate(self.bc.orderedvals()):
                if isinstance(bci, sp.Expr):
                    X = T.local_mesh(True)
                    for sym in bci.free_symbols:
                        tt = sp.symbols('t', real=True)
                        if sym == tt:
                            bci = bci.subs(tt, self.bc_time)
                    sym0 = bci.free_symbols
                    lbci = sp.lambdify(tuple(sym0), bci, 'numpy')
                    Yi = []
                    for sym in sym0:
                        k = 'xyzrs'.index(str(sym))
                        Yi.append(X[k][this_base.si[j]])
                    f_bci = lbci(*Yi)
                    # Put the value in the position of the bc dofs
                    if s.stop == int(this_base.N*this_base.padding_factor):
                        b[this_base.si[-num_bcs+j]] = f_bci

                elif isinstance(bci, Function):
                    # if bci is some Function that needs to be evaluated at the boundary
                    sl = list(T.local_slice(False))
                    sl.pop(self.axis)
                    sl = tuple(sl)
                    gs = list(T.global_shape())
                    gs.pop(self.axis)
                    gs = tuple(gs)
                    mesh = []
                    sj = this_base.si[0]
                    for axis, mj in enumerate(T.mesh(True)):
                        mj = np.broadcast_to(mj[sj], gs)
                        if axis != self.axis:
                            mesh.append(mj.flatten())
                    name = self.bc.orderednames()[j]
                    dm = this_base.domain[0] if 'L' in name else this_base.domain[1]
                    mesh.insert(self.axis, np.ones(len(mesh[0]))*dm)
                    xj = np.vstack(mesh)
                    newb = bci.function_space().eval(xj, bci).reshape(gs)
                    if s.stop == int(this_base.N*this_base.padding_factor):
                        b[this_base.si[-num_bcs+j]] = newb[sl]

                elif isinstance(bci, (Number, np.ndarray)):
                    if s.stop == int(this_base.N*this_base.padding_factor):
                        b[this_base.si[-num_bcs+j]] = bci

                else:
                    raise NotImplementedError

            if len(T.get_nonhomogeneous_axes()) == 1:
                for j, bci in enumerate(self.bc.orderedvals()):
                    from .spectralbase import FuncWrap
                    if number_of_bases_after_this == 0:
                        # Inhomogeneous base is the first to be transformed
                        b_hat = b
                    else:
                        i = number_of_bases_after_this
                        u_hat = T.forward._xfftn[i].input_array
                        dx = T.coors.get_sqrt_det_g()
                        msdict = split(dx)
                        c = msdict[0]['coeff']
                        fun = FuncWrap(lambda x, y, **kw: x, u_hat, u_hat) # Do-nothing
                        fwd = Transform(list(T.forward._xfftn[:i])+[fun],
                                        T.forward._transfer[:i], [[], []])
                        b_hat = fwd(c*b).copy()

                    # Now b_hat contains the correct slices in slm1 and slm2
                    # These are the values to use on intermediate steps.
                    # If for example a Dirichlet space is squeezed between two Fourier spaces
                    self.bcs[j] = b_hat[this_base.si[-num_bcs+j]].copy()
                b_hat = T.forward(b).copy()

                for j in range(self.bc.num_bcs()):
                    self.bcs_final[j] = b_hat[this_base.si[-num_bcs+j]].copy()

            elif len(T.get_nonhomogeneous_axes()) == 2:
                assert len(T.bases) == 2, 'Only implemented for 2D'
                bases = []
                for axis, base in enumerate(T.bases):
                    if not base is this_base:
                        bases.append(base.get_unplanned())

                other_base = bases[0]
                ua = Array(other_base)
                lr = {'left': 0, 'right': 1}
                num_bcs = this_base.bcs.num_bcs()
                df = 2./(other_base.domain[1]-other_base.domain[0])
                for i, bcj in enumerate(this_base.bcs.orderedvals()):
                    sym = sp.sympify(bcj).free_symbols
                    xx = sym.pop() if len(sym) == 1 else None
                    for lr_other, bco in other_base.bcs.items():
                        xj = other_base.domain[lr[lr_other]]
                        for key in bco.keys():
                            if xx is not None:
                                if key == 'D':
                                    bco[key] = bcj.subs(xx, xj)
                                elif key == 'N':
                                    bco[key] = bcj.diff(xx, 1).subs(xx, xj)/df
                                elif key == 'N2':
                                    bco[key] = bcj.diff(xx, 2).subs(xx, xj)/df**2
                                elif key == 'N3':
                                    bco[key] = bcj.diff(xx, 3).subs(xx, xj)/df**3
                                elif key == 'N4':
                                    bco[key] = bcj.diff(xx, 4).subs(xx, xj)/df**4
                            else:
                                bco[key] = bcj
                        other_base.bc.bcs[:] = other_base.bcs.orderedvals()
                        other_base.bc.bcs_final[:] = other_base.bcs.orderedvals()
                    ua[:] = b[this_base.si[-num_bcs+i]]
                    self.bcs_final[i] = project(ua, other_base)

    def _add_to_orthogonal(self, u, uh):
        """Add contribution from boundary functions to ``u``

        Parameters
        ----------
        u : Function
            Apply boundary values to this Function
        uh : Function
            Containing correct boundary values of Function
        """
        B = self.base.get_bc_space()
        M = (B.stencil_matrix().T).astype(float)
        sl = B.slice()
        ds = uh.shape[self.base.axis] - B.N
        if ds > 0: # uh is padded, but B is not. Boundary dofs are in padded locations.
            sl = slice(sl.start + ds, sl.stop + ds)
        if u.ndim > 1:
            sl = self.base.sl[sl]
        for i, row in enumerate(M):
            u[self.base.si[i]] += np.sum(self.base.broadcast_to_ndims(row)*uh[sl], axis=self.base.axis)

    def _add_mass_rhs(self, u):
        """Add contribution of boundary functions to rhs before solving in forward transforms

        Parameters
        ----------
        u : Function
            Add boundary values to this Function

        """
        self.set_boundary_dofs(u)
        if not self.has_nonhomogeneous_bcs(): # there's no contribution for homogeneous bcs
            return
        coors = self.tensorproductspace.coors if self.tensorproductspace else self.base.coors
        M = self.base.get_bcmass_matrix(coors.get_sqrt_det_g())
        w0 = np.zeros_like(u)
        u -= M.matvec(u, w0, axis=self.base.axis)

    def set_boundary_dofs(self, u, final=False):
        """Apply boundary condition after solving, fixing boundary dofs

        Parameters
        ----------
        u : Function
            Apply boundary values to this Function
        final : bool
            Whether the function is fully transformed or not. Not is used in
            forward transforms, where the transform of this base may be in
            between the transforms of other bases. If final is True, then u
            must be a fully transformed Function.

        """
        M = self.bc.num_bcs()
        if final is True:
            for i in range(M):
                u[self.base.si[-(M)+i]] = self.bcs_final[i]
        else:
            for i in range(M):
                u[self.base.si[-(M)+i]] = self.bcs[i]

    def has_nonhomogeneous_bcs(self):
        for bc in self.bc.orderedvals():
            if isinstance(bc, Number):
                if not bc == 0:
                    return True
            elif isinstance(bc, np.ndarray):
                if not np.all(bc == 0):
                    return True
            elif isinstance(bc, sp.Expr):
                return True
        return False



import importlib
from copy import copy
from scipy.special import gammaln
import numpy as np
from numpy.polynomial import chebyshev as n_cheb
from mpi4py_fft import fftw
from mpi4py import MPI
from mpi4py_fft.fftw.utilities import FFTW_MEASURE, FFTW_PRESERVE_INPUT
from shenfun.optimization import runtimeoptimizer
from shenfun.optimization.cython import Leg2Cheb, Cheb2Leg, omega
from shenfun.spectralbase import islicedict, slicedict
from shenfun.forms.arguments import FunctionSpace
from shenfun.forms import project
from . import fastgl

__all__ = ['DLT', 'leg2cheb', 'cheb2leg', 'Leg2chebHaleTownsend',
           'Leg2Cheb', 'Cheb2Leg', 'FMMLeg2Cheb', 'FMMCheb2Leg']


class DLT:
    r"""Discrete Legendre Transform

    A class for performing fast FFT-based discrete Legendre transforms, both
    forwards and backwards. Based on::

        Nicholas Hale and Alex Townsend "A fast FFT-based discrete Legendre
        transform", IMA Journal of Numerical Analysis (2015)
        (https://arxiv.org/abs/1505.00354)

    Parameters
    ----------
    input_array : real or complex array
    s : sequence of ints, optional
        Not used - included for compatibility with Numpy
    axes : integer or 1-tuple of int, optional
        Axis over which to compute the DLT. Named axes for compatibility.
    threads : int, optional
        Number of threads used in computing DLT.
    kind : str, optional
        Either one of

            - 'forward'
            - 'backward'
            - 'scalar product'

        The scalar product is exactly like the forward transform, except that
        the Legendre mass matrix is not applied to the output.
    flags : sequence of ints, optional
        Flags from

            - FFTW_MEASURE
            - FFTW_EXHAUSTIVE
            - FFTW_PATIENT
            - FFTW_DESTROY_INPUT
            - FFTW_PRESERVE_INPUT
            - FFTW_UNALIGNED
            - FFTW_CONSERVE_MEMORY
            - FFTW_ESTIMATE
    output_array : complex array, optional
        Array to be used as output array. Must be of correct shape, type,
        strides and alignment

    Note
    ----
    The Legendre series expansion of :math:`f(x)` is

    .. math::

        f(x) = \sum_{k=0}^{N-1} \hat{f}_k L_k(x)

    The series evaluated at quadrature points :math:`\{x_j\}_{j=0}^{N-1}` gives the
    vector :math:`\{f(x_j)\}_{j=0}^{N-1}`. We define a forward transform as
    computing :math:`\{\hat{f}_k\}_{k=0}^{N-1}` from :math:`\{f(x_j)\}_{j=0}^{N-1}`.
    The other way around, a backward transform, computes :math:`\{f(x_j)\}_{j=0}^{N-1}`
    given :math:`\{\hat{f}_k\}_{k=0}^{N-1}`. This is in agreement with shenfun's
    definition of forward/backward directions, but it disagrees with the definitions
    used by Hale and Townsend.

    Also note that this `fast` transform is actually slower than the default recursive
    version for approximately :math:`N<1000`.

    Example
    -------
    >>> import numpy as np
    >>> from shenfun import legendre, FunctionSpace, Function, Array
    >>> N = 4
    >>> L = FunctionSpace(N, 'L')
    >>> u = Function(L)
    >>> u[1] = 1
    >>> c = Array(L)
    >>> c = L.backward(u, c, kind='fast')
    >>> print(c)
    [-0.86113631 -0.33998104  0.33998104  0.86113631]
    >>> u[:] = 1
    >>> np.alltrue(np.abs(u - u.backward().forward()) < 1e-12)
    True

    """
    def __init__(self, input_array, s=None, axes=(-1,), threads=1, kind='backward',
                 flags=(FFTW_MEASURE, FFTW_PRESERVE_INPUT), output_array=None):
        if isinstance(axes, tuple):
            assert len(axes) == 1
            axis = axes[-1]
        elif isinstance(axes, int):
            axis = axes
        self.axis = axis
        assert kind in ('forward', 'scalar product', 'backward')
        self.kind = kind
        self.sl = slicedict(axis=axis, dimensions=input_array.ndim)
        self.si = islicedict(axis=axis, dimensions=input_array.ndim)
        N = self.N = input_array.shape[axis]
        xl, wl = fastgl.leggauss(N)
        xc = n_cheb.chebgauss(N)[0]
        thetal = np.arccos(xl)[::-1]
        thetac = np.arccos(xc)
        s = [None]*input_array.ndim
        s[axis] = slice(None)
        s = tuple(s) # broadcast to ndims
        self.dtheta = (thetal - thetac)[s]
        self.n = np.arange(N, dtype=float)[s]
        if kind in ('forward', 'scalar product'):
            self.wl = wl[s]
            self.nsign = np.ones(N)
            self.nsign[1::2] = -1
            if kind == 'forward': # apply inverse mass matrix as well
                self.nsign = self.nsign[s]*(self.n+0.5)
            else:
                self.nsign = self.nsign[s]
        ck = np.full(N, np.pi/2); ck[0] = np.pi
        self.ck = ck[s]
        U = input_array
        V = output_array if output_array is not None else U.copy()
        self.plan(U, V, kind, threads, flags)
        ##self.leg2chebclass = Leg2Cheb(U, axis=axis, maxs=100, use_direct=500)
        self.leg2chebclass = Leg2Cheb(U, domains=2, diagonals=16, axis=axis, maxs=100, use_direct=1000)

    def plan(self, U, V, kind, threads, flags):
        Uc = U.copy()
        Vc = V.copy()
        # dct and dst use the same input/output arrays
        if kind in ('forward', 'scalar product'):
            self.dct = DCT(Uc, axis=self.axis, type=2, threads=threads, flags=flags, output_array=Vc)
            self.dst = DST(Uc, axis=self.axis, type=2, threads=threads, flags=flags, output_array=Vc)
        else:
            self.dct = DCT(Uc, axis=self.axis, type=3, threads=threads, flags=flags, output_array=Vc)
            self.dst = DST(Uc, axis=self.axis, type=3, threads=threads, flags=flags, output_array=Vc)
        self._input_array = U
        self._output_array = V

    @property
    def input_array(self):
        return self._input_array

    @property
    def output_array(self):
        return self._output_array

    def __call__(self, input_array=None, output_array=None, **kw):
        if input_array is not None:
            self._input_array[:] = input_array

        # Set up x, which is input array to dct/dst
        # Some dcts apparently destroys the input_array, so need copy
        x = np.zeros_like(self.input_array)
        x[:] = self._input_array
        if self.kind in ('forward', 'scalar product'):
            x *= self.wl
        else:
            x = self.leg2chebclass(x.copy(), x)

        fk = self.dct(x).copy()
        nfac = 1
        y = 1
        n = 1
        converged = False
        while not converged:
            even = n % 2 == 0
            deven = (n+1)//2 % 2 == 0
            nfac *= n
            if self.kind == 'backward':
                x *= self.n
                y *= self.dtheta
            else:
                x *= self.dtheta
                y *= self.n
            sign = 1 if deven else -1
            fft = self.dct if even else self.dst
            h = fft(x)
            df = sign/nfac*y*h
            fk += df
            error = np.linalg.norm(df)
            #print(f"{n:4d} {error:2.4e}")
            converged = error < 1e-16
            n += 1

        if self.kind in ('forward', 'scalar product'):
            fk = self.leg2chebclass(fk.copy(), fk, transpose=True)
            fk *= self.nsign
        else:
            fk[:] = fk[self.sl[slice(-1, None, -1)]] # reverse
        if output_array is not None:
            output_array[:] = fk
            return output_array
        self._output_array[:] = fk
        return self._output_array

class DCT:
    """Discrete cosine transform with appropriate scaling

    input_array : array
    s : sequence of ints, optional
        Not used - included for compatibility with Numpy
    axes : sequence of ints, optional
        Axes over which to compute the real-to-real dct.
    type : int, optional
        Type of `dct <http://www.fftw.org/fftw3_doc/Real_002dto_002dReal-Transform-Kinds.html>`_

            - 1 - FFTW_REDFT00
            - 2 - FFTW_REDFT10,
            - 3 - FFTW_REDFT01,
            - 4 - FFTW_REDFT11
    threads : int, optional
        Number of threads used in computing dct.
    flags : sequence of ints, optional
        Flags from

            - FFTW_MEASURE
            - FFTW_EXHAUSTIVE
            - FFTW_PATIENT
            - FFTW_DESTROY_INPUT
            - FFTW_PRESERVE_INPUT
            - FFTW_UNALIGNED
            - FFTW_CONSERVE_MEMORY
            - FFTW_ESTIMATE
    output_array : array, optional
        Array to be used as output array. Must be of correct shape, type,
        strides and alignment

    """
    def __init__(self, input_array, type=3, s=None, axis=-1, threads=1,
                 flags=(FFTW_MEASURE,), output_array=None):
        self.axis = axis
        self.type = type
        self.dct = fftw.dctn(input_array, axes=(axis,), type=type, threads=threads,
                             flags=flags, output_array=output_array)
        s = [slice(None)]*input_array.ndim
        s[self.axis] = slice(0, 1)
        self.s = tuple(s)

    @property
    def input_array(self):
        return self.dct.input_array

    @property
    def output_array(self):
        return self.dct.output_array

    def __call__(self, input_array=None, output_array=None):
        if input_array is not None:
            self.input_array[:] = input_array
        out = self.dct()
        if self.type == 3:
            out += self.dct.input_array[self.s]
        out /= 2
        if output_array is not None:
            output_array[:] = out
            return output_array
        return out

class DST:
    """Discrete sine transform with appropriate scaling

    Parameters
    ----------
    input_array : array
    s : sequence of ints, optional
        Not used - included for compatibility with Numpy
    axes : sequence of ints, optional
        Axes over which to compute the real-to-real dst.
    type : int, optional
        Type of `dst <http://www.fftw.org/fftw3_doc/Real_002dto_002dReal-Transform-Kinds.html>`_

            - 1 - FFTW_RODFT00
            - 2 - FFTW_RODFT10
            - 3 - FFTW_RODFT01
            - 4 - FFTW_RODFT11
    threads : int, optional
        Number of threads used in computing dst.
    flags : sequence of ints, optional
        Flags from

            - FFTW_MEASURE
            - FFTW_EXHAUSTIVE
            - FFTW_PATIENT
            - FFTW_DESTROY_INPUT
            - FFTW_PRESERVE_INPUT
            - FFTW_UNALIGNED
            - FFTW_CONSERVE_MEMORY
            - FFTW_ESTIMATE
    output_array : array, optional
        Array to be used as output array. Must be of correct shape, type,
        strides and alignment

    """
    def __init__(self, input_array, type=3, s=None, axis=-1, threads=1,
                 flags=None, output_array=None):
        self.axis = axis
        self.type = type
        self.dst = fftw.dstn(input_array, axes=(axis,), type=type, threads=threads,
                             flags=flags, output_array=output_array)
        self.sl = slicedict(axis=axis, dimensions=input_array.ndim)
        self.si = islicedict(axis=axis, dimensions=input_array.ndim)

    @property
    def input_array(self):
        return self.dst.input_array

    @property
    def output_array(self):
        return self.dst.output_array

    def destroy(self):
        self.dst.destroy()

    def __call__(self, input_array=None, output_array=None):
        if input_array is not None:
            self.input_array[:] = input_array
        x = self.input_array
        if self.type == 3:
            x[self.sl[slice(0, -1)]] = x[self.sl[slice(1, None)]]
            x[self.si[-1]] = 0
            out = self.dst()
        elif self.type == 2:
            out = self.dst()
            out[self.sl[slice(1, None)]] = out[self.sl[slice(0, -1)]]
            out[self.si[0]] = 0
        out /= 2
        if output_array is not None:
            output_array[:] = out
            return output_array
        return out

#Omega = lambda z: np.exp(gammaln(z+0.5) - gammaln(z+1))
Omega = omega

LxyE = lambda x, y: -0.5/(2*x+2*y+1)/(y-x)*Omega(y-x-1)*Omega(y+x-0.5)
MxyE = lambda x, y: Omega(y-x)*Omega(y+x)
Mxy = lambda x, y: Omega((y-x)/2)*Omega((y+x)/2)
Lxy = lambda x, y: -1/(x+y+1)/(y-x)*Omega((y-x-2)/2)*Omega((y+x-1)/2)
Hxy = lambda x, y: Omega((y-x)/2)*Omega((y+x)/2)/np.sqrt((x+y)*(y-x))

@runtimeoptimizer
def leg2cheb(cl, cc=None, axis=0, transpose=False):
    r"""Compute Chebyshev coefficients from Legendre coefficients

    .. math::

            \hat{c}^{cheb} = M \hat{c}^{leg}

    where :math:`\hat{c}^{cheb} \in \mathbb{R}^N` are the Chebyshev
    coefficients, :math:`\hat{c}^{leg} \in \mathbb{R}^N` the Legendre
    coefficients and :math:`M\in\mathbb{R}^{N \times N}` the matrix for the
    conversion. Note that if keyword transpose is true, then we compute

    .. math::

        \hat{a} = M^T \hat{b}

    for some vectors :math:`\hat{a}` and :math:`\hat{b}`.

    Paramters
    ---------
    cl : array
        The Legendre coefficients (if transpose is False)
    cc : array (return array), optional
        The Chebyshev coefficients (if transpose is False)
    axis : int, optional
        The axis over which to take the transform
    transpose : bool
        Whether to perform the transpose operation

    Note
    ----
    This is a low-memory implementation of the direct matrix vector
    product. The matrix :math:`M` is given by Alpert and Rokhlin 'A Fast Algorithm
    for the Evaluation of Legendre Expansions', SIAM Journal on Scientific
    and Statistical Computing, 12, 1, 158-179, (1991), 10.1137/0912009.
    The matrix is not explicitely created.
    """
    if cc is None:
        cc = np.zeros_like(cl)
    else:
        cc.fill(0)

    if axis > 0:
        cl = np.moveaxis(cl, axis, 0)
        cc = np.moveaxis(cc, axis, 0)
    N = cl.shape[0]
    k = np.arange(N)
    a = Omega(k)
    sl = [None]*cc.ndim
    ck = np.full(N, np.pi/2); ck[0] = np.pi
    sl[0] = slice(None)
    ck = ck[tuple(sl)]
    if transpose is False:
        for n in range(0, N, 2):
            sl[0] = slice(n//2, N-n//2)
            cc[:(N-n)] += a[n//2]*a[tuple(sl)]*cl[n:]
        cc /= ck

    else:
        cx = cl/ck
        for n in range(0, N, 2):
            sl[0] = slice(n//2, N-n//2)
            cc[n:] += a[n//2]*a[tuple(sl)]*cx[:(N-n)]

    if axis > 0:
        cl = np.moveaxis(cl, 0, axis)
        cc = np.moveaxis(cc, 0, axis)
    return cc

@runtimeoptimizer
def cheb2leg(cc, cl=None, axis=0):
    r"""Compute Legendre coefficients from Chebyshev coefficients

    .. math::

            \hat{c}^{leg} = L \hat{c}^{cheb}

    where :math:`\hat{c}^{cheb} \in \mathbb{R}^N` are the Chebyshev
    coefficients, :math:`\hat{c}^{leg} \in \mathbb{R}^N` the Legendre
    coefficients and :math:`L\in\mathbb{R}^{N \times N}` the matrix for the
    conversion.

    Paramters
    ---------
    cc : array
        The Chebyshev coefficients
    cl : array
        The Legendre coefficients (return array)
    axis : int, optional
        The axis over which to take the transform

    Note
    ----
    This is a low-memory implementation of the direct matrix vector
    product. The matrix :math:`L` is given by Alpert and Rokhlin 'A Fast Algorithm
    for the Evaluation of Legendre Expansions', SIAM Journal on Scientific
    and Statistical Computing, 12, 1, 158-179, (1991), 10.1137/0912009.
    The matrix is not explicitely created.
    """
    if cl is None:
        cl = np.zeros_like(cc)
    else:
        cl.fill(0)
    if axis > 0:
        cl = np.moveaxis(cl, axis, 0)
        cc = np.moveaxis(cc, axis, 0)
    N = cc.shape[0]
    k = np.arange(N, dtype=float)
    k[0] = 1
    vn = cc*k
    a = 1/(2*Omega(k)*k*(k+0.5))
    k[0] = 0
    a[0] = 2/np.sqrt(np.pi)
    cl[:] = np.sqrt(np.pi)*a*vn
    for n in range(2, N, 2):
        dn = Omega(np.array([(n-2)/2]))/n
        cl[:(N-n)] -= dn*a[n//2:(N-n//2)]*vn[n:]
    cl *= (k+0.5)
    if axis > 0:
        cl = np.moveaxis(cl, 0, axis)
        cc = np.moveaxis(cc, 0, axis)
    return cl

class Leg2chebHaleTownsend:
    """Class for computing Chebyshev coefficients from Legendre coefficients

    Algorithm from::

        Nicholas Hale and Alex Townsend "A fast, simple and stable Chebyshev-
        Legendre transform using an asymptotic formula", SIAM J Sci Comput (2014)
        (https://epubs.siam.org/doi/pdf/10.1137/130932223)

    Parameters
    ----------
    input_array : array
        Legendre coefficients
    output_array : array
        The returned array
    axis : int
        The axis over which to perform the computation in case the input_array
        is multidimensional.
    nM : int
        Parameter, see Hale and Townsend (2014). Note that one must have N >> nM.
    Nmin : int
        Parameter. Choose direct matvec approach for N < Nmin
    """
    def __init__(self, input_array, output_array=None, axis=0, nM=50, Nmin=40000):
        self.axis = axis
        self.N = input_array.shape[axis]
        self.L = None
        self.T = None
        self.U = None
        self.a = None
        self.Nmin = Nmin
        self._input_array = input_array
        self._output_array = output_array if output_array is not None else input_array.copy()
        self.sl = slicedict(axis=axis, dimensions=input_array.ndim)
        if self.N > Nmin:
            from shenfun import config
            mod = config['optimization']['mode']
            self.lib = importlib.import_module('.'.join(('shenfun.optimization', mod, 'transforms')))
            N = self.N
            self.thetak = (np.arange(N)+0.5)*np.pi/N
            self.sintk = np.sin(self.thetak)
            alpha = self.alpha = min(1/np.log(N/nM), 0.5)
            if N < 1000:
                K = 1
            elif N < 10000:
                K = 2
            elif N < 1000000:
                K = 3
            else:
                K = 4
            self.K = K
            self.ix = {0: 0}
            for k in range(1, K+1):
                self.ix[k] = int((N + 1) / np.pi * np.arcsin(nM / N / alpha**k))

    @property
    def input_array(self):
        return self._input_array

    @property
    def output_array(self):
        return self._output_array

    def _Um(self, m):
        return np.sin((m+0.5)*(np.pi/2-self.thetak))/(2**(m+0.5)*self.sintk**(m-0.5))

    def _Vm(self, m):
        return np.cos((m+0.5)*(np.pi/2-self.thetak))/(2*self.sintk)**(m+0.5)

    @staticmethod
    def _Cn(n):
        return np.sqrt(4/np.pi)*np.exp(gammaln(n+1) - gammaln(n+3/2))

    def plan(self, input_array):
        from shenfun import FunctionSpace
        assert input_array.shape[self.axis] == self.N
        if self.T is not None:
            assert self.T.N == self.N
            return

        self.L = FunctionSpace(self.N, 'L')
        self.U = FunctionSpace(self.N, 'U', quad='GC')
        self.T = FunctionSpace(self.N, 'C', quad='GC')
        self.a = self.L.get_recursion_matrix(self.N+3, self.N+3).diags('dia').data

        if input_array.ndim > 1:
            self.T.plan(input_array.shape, self.axis, input_array.dtype, {})
            self.U.plan(input_array.shape, self.axis, input_array.dtype, {})

    def __call__(self, input_array=None, output_array=None, transpose=False):
        r"""Compute Chebyshev coefficients from Legendre. That is, compute

        .. math::

            \hat{c}^{cheb} = M \hat{c}^{leg}

        where :math:`\hat{c}^{cheb} \in \mathbb{R}^N` are the Chebyshev
        coefficients, :math:`\hat{c}^{leg} \in \mathbb{R}^N` the Legendre
        coefficients and :math:`M\in\mathbb{R}^{N \times N}` the matrix for the
        conversion. Note that if keyword 'transpose' is true, then we compute

        .. math::

            \hat{a} = M^T \hat{b}

        for some vectors :math:`\hat{a}` and :math:`\hat{b}`.

        The Chebyshev and Legendre coefficients are the regular coefficients to
        the series

        .. math::

            p_l(x) = \sum_{k=0}^{N} \hat{c}_k^{leg}L_k(x) \\
            p_c(x) = \sum_{k=0}^{N} \hat{c}_k^{cheb}T_k(x)

        and we get :math:`\{\hat{c}_k^{cheb}\}_{k=0}^N` by setting :math:`p_l(x)=p_c(x)`
        for :math:`x=\{x_i\}_{i=0}^N`, where :math:`x_i=\cos(i+0.5)\pi/N`.

        Parameters
        ----------
        input_array : array
        output_array : array
        transpose : bool
            Whether to compute the transpose of the regular transform

        Note
        ----
        For small N we use a direct method that costs approximately :math:`0.25 N^2`
        operations. For larger N (see 'Nmin' parameter) we use the fast routine of

            Hale and Townsend 'A fast, simple and stable Chebyshev-Legendre
            transform using an asymptotic formula', SIAM J Sci Comput (2014)

        """
        if input_array is not None:
            self._input_array[:] = input_array

        axis = self.axis
        N = self.N
        if N <= self.Nmin:
            out = leg2cheb(self.input_array, self.output_array, axis=axis, transpose=transpose)
            if output_array is not None:
                output_array[:] = out
                return output_array
            return out

        self.plan(self.input_array)
        sn = [None]*self.input_array.ndim
        sn[axis] = slice(None); sn = tuple(sn)
        hmn = np.ones(self.N)
        Tc = np.zeros_like(self.input_array)
        Uc = np.zeros_like(self.input_array)
        z = np.zeros_like(self.input_array)
        cn = self._Cn(np.arange(N))[sn]
        xi, wi = self.T.points_and_weights()

        if transpose is False:
            cn = self.input_array*cn
            na = np.arange(N)
            for m in range(10):
                if m > 0:
                    hmn *= (m-0.5)**2/(m*(na+m+0.5))
                cm = cn*hmn[sn]
                um = self._Um(m)[sn]
                vm = self._Vm(m)[sn]
                for k in range(1, self.K+1):
                    Tc[:] = 0
                    Uc[:] = 0
                    si = self.sl[slice(self.ix[k], N-self.ix[k])]
                    sk = self.sl[slice(int(self.alpha**k*N), int(self.alpha**(k-1)*N))]
                    Tc[sk] = cm[sk]
                    Uc[self.sl[slice(0, -1)]] = Tc[self.sl[slice(1, None)]]
                    z[si] += (vm*self.T.backward(Tc) + um*self.U.backward(Uc))[si]
            si = self.sl[slice(0, self.ix[1])]
            zx = np.zeros_like(z[si])
            self.lib.evaluate_expansion_all(self.input_array, zx, xi[si[axis]], self.axis, self.a) # recursive eval
            z[si] += zx
            si = self.sl[slice(N-self.ix[1], None)]
            self.lib.evaluate_expansion_all(self.input_array, zx, xi[si[axis]], self.axis, self.a)
            z[si] += zx
            for k in range(1, self.K):
                si = self.sl[slice(self.ix[k], self.ix[k+1])]
                sk = self.sl[slice(0, int(self.alpha**k*self.N))]
                zx = np.zeros_like(z[si])
                self.lib.evaluate_expansion_all(self.input_array[sk], zx, xi[si[axis]], axis, self.a)
                z[si] += zx
                si = self.sl[slice(self.N-self.ix[k+1], self.N-self.ix[k])]
                self.lib.evaluate_expansion_all(self.input_array[sk], zx, xi[si[axis]], axis, self.a)
                z[si] += zx
            si = self.sl[slice(self.ix[self.K], self.N-self.ix[self.K])]
            sk = self.sl[slice(0, int(self.alpha**(self.K)*self.N))]
            zx = np.zeros_like(z[si])
            self.lib.evaluate_expansion_all(self.input_array[sk], zx, xi[si[axis]], axis, self.a)
            z[si] += zx
            self._output_array = self.T.forward(z, self._output_array)

        else: # transpose
            ck = np.full(N, np.pi/2); ck[0] = np.pi
            ctilde = self.T.backward(self.input_array/ck[sn]).copy() # TN^{-T} * cl
            wu = self.U.points_and_weights()[1]
            U1 = np.zeros_like(self.input_array)
            for m in range(10):
                if m > 0:
                    hmn *= (m-0.5)**2/(m*(np.arange(N)+m+0.5))
                um = self._Um(m)[sn]
                vm = self._Vm(m)[sn]
                for k in range(1, self.K+1):
                    Tc[:] = 0
                    Uc[:] = 0
                    si = self.sl[slice(self.ix[k], self.N-self.ix[k])]
                    sk = self.sl[slice(int(self.alpha**k*self.N), int(self.alpha**(k-1)*self.N))]
                    Tc[si] = ctilde[si]
                    T0 = self.T.scalar_product(Tc*vm)
                    U0 = self.U.scalar_product(Tc/wu[sn]*wi[sn]*um)
                    U1[self.sl[slice(1, None)]] = U0[self.sl[slice(0, -1)]]
                    z[sk] += (cn*hmn[sn]*(T0+U1))[sk]
            sk = self.sl[slice(0, int(self.alpha**self.K*self.N))]
            zx = np.zeros_like(z[sk])
            z[sk] += restricted_product(self.L, wi[sn]*ctilde, zx, xi, 0, self.N, 0, axis, self.a)
            for k in range(self.K):
                sk = self.sl[slice(int(self.alpha**(k+1)*self.N), int(self.alpha**k*self.N))]
                zx = np.zeros_like(z[sk])
                z[sk] += restricted_product(self.L, wi[sn]*ctilde, zx, xi, 0, self.ix[k+1], sk[axis].start, axis, self.a)
                z[sk] += restricted_product(self.L, wi[sn]*ctilde, zx, xi, N-self.ix[k+1], N, sk[axis].start, axis, self.a)
            self._output_array[:] = z

        if output_array is not None:
            output_array[:] = self._output_array
            return output_array
        return self._output_array

@runtimeoptimizer
def restricted_product(L, input_array, output_array, xi, i0, i1, a0, axis, a):
    r"""Returns the restricted matrix vector product

    .. math::

        \sum_{i=i_0}^{i_1} \sum_{k=k_0}^{k_1} \hat{u}_k P_k(x_i)

    where the matrix :math:`P_k(x_i)` is the :math:`k`'th basis function of a
    family of orthogonal polynomials, at points :math:`\{x_i\}_{i_0}^{i_1}`.
    The array :math:`\boldsymbol{\hat{u}}` can be multidimensional, in which
    case the product is applied along the indicated axis.

    Parameters
    ----------
    L : instance of :class:`.SpectralBase`
    input_array : array
    output_array : array
    xi : Quadrature points for space L
    i0, i1 : integers
        Start and stop indices for xi
    a0 : int
        Start index for basis functions :math:`P_k`
    axis : int
        The axis of :math:`\hat{u}_k` to take the restricted product over in
        case :math:`\boldsymbol{\hat{u}}` is multidimensional
    a : array
        Recurrence coefficients for the orthogonal polynomials

    """
    N = xi.shape[0]
    Lnm = L.evaluate_basis(xi[i0:i1], i=a0)
    Ln = L.evaluate_basis(xi[i0:i1], i=a0+1)
    if a.shape[0] == 3:
        anm = a[0] # a_{n-1, n}
        ann = a[1] # a_{n, n}
        anp = a[2] # a_{n+1, n}
    else:
        anm = a[0]
        anp = a[1]
        ann = np.zeros(N+2)
    Lnp = (xi[i0:i1]-ann[1+a0])/anm[1+a0]*Ln - anp[1+a0]/anm[1+a0]*Lnm
    for k in range(len(output_array)):
        kp = k+a0
        s1 = 1/anm[kp+2]
        s2 = anp[kp+2]/anm[kp+2]
        a00 = ann[kp+2]
        s = 0.0
        for i in range(len(Ln)):
            s += Lnm[i]*input_array[i0+i]
            Lnm[i] = Ln[i]
            Ln[i] = Lnp[i]
            Lnp[i] = s1*(xi[i0+i]-a00)*Ln[i] - s2*Lnm[i]
        output_array[k] = s
    return output_array

def getChebyshev(level, D, s, diags, A, N, l2c=True):
    """Low-rank computation of Chebyshev coefficients

    Computes Chebyshev coefficients for all submatrices on a given
    level of a hierarchical decomposition.

    Parameters
    ----------
    level : int
        The level in the hierarchical decomposition
    D : int
        Domains size on level
    s : int
        Size of the smallest submatrices
    diags : int
        The number of neglected diagonals, that are treated using a
        direct approach
    A : list
        Each item is a flattened low-rank matrix of coefficients
    N : list
        Each item is the shape of the corresponding matrix item in A
    l2c : bool
        If True, the transform goes from Legendre to Chebyshev, and
        vice versa if False
    """
    from shenfun import FunctionSpace, TensorProductSpace
    h = s*get_h(level, D)
    i0, j0 = get_ij(level, 0, s, D, diags)
    Nb = get_number_of_blocks(level, D)
    T0 = FunctionSpace(100, 'C', domain=[j0+2*h, j0+4*h])
    fun = Mxy if l2c == True else Lxy
    w0 = fun(2*h-1, T0.mesh())
    m0 = T0.forward(w0)
    z = np.where(np.diff(abs(m0)) > 0)[0]
    Nx = 100 if len(z) < 2 else z[1]
    T0.domain = [j0+4*h, j0+6*h]
    w0 = fun(2*h-1, T0.mesh())
    m0 = T0.forward(w0)
    z = np.where(np.diff(abs(m0)) > 0)[0]
    Nx2 = 100 if len(z) < 2 else z[1]
    Nx2 = min(Nx2, Nx)
    #Nx, Nx2 = 50, 50
    T1 = FunctionSpace(Nx, 'C')
    S = TensorProductSpace(MPI.COMM_SELF, (T1, T1))
    f = np.zeros((Nx, Nx))
    dctn = fftw.dctn(np.zeros((Nx, Nx)), axes=(0, 1))
    xj = np.cos((np.arange(Nx)+0.5)*np.pi/Nx)
    xj2 = np.cos((np.arange(Nx2)+0.5)*np.pi/(Nx2))
    dctn2 = fftw.dctn(np.zeros((Nx2, Nx2)), axes=(0, 1))
    X = np.zeros((Nx, 1))
    Y = np.zeros((1, Nx))
    X2 = np.zeros((Nx2, 1))
    Y2 = np.zeros((1, Nx2))
    Mmax = 0
    for k in range(Nb):
        i0, j0 = get_ij(level, k, s, D, diags)
        for j in range(D[level]):
            S.bases[1].domain = 2*(j0+(j+1)*h), 2*(j0+(j+2)*h)
            for i in range(j+1):
                S.bases[0].domain = 2*(i0+i*h), 2*(i0+(i+1)*h)
                f[:] = 0
                if j > i:
                    Y2[0, :] = S.bases[1].map_true_domain(xj2)
                    X2[:, 0] = S.bases[0].map_true_domain(xj2)
                    f[:Nx2, :Nx2] = dctn2(fun(X2, Y2))/Nx2**2
                else:
                    Y[0, :] = S.bases[1].map_true_domain(xj)
                    X[:, 0] = S.bases[0].map_true_domain(xj)
                    f[:] = dctn(fun(X, Y))/Nx**2
                f[0] /= 2
                f[:, 0] /= 2
                z = np.where(np.diff(abs(f[0, 2:])) >= 0)[0]
                #z = np.where(abs(f[0]) < 1e-15)[0]
                Mmin = z[0]+2 if len(z) > 1 else Nx
                #Mmin = Nx
                A.append(f[:Mmin, :Mmin].ravel().copy())
                N.append(Mmin)
                Mmax = max(Mmax, Mmin)
    S.destroy()
    return Mmax

class FMMLevel:
    """Abstract base class for hierarchical matrix
    """
    def __init__(self, N, diagonals=8, domains=None, levels=None, l2c=True, maxs=100, use_direct=-1):
        self.N = N
        self.use_direct = use_direct
        self._output_array = np.array([0])
        if N <= use_direct:
            self.Nn = N
            return

        if domains is not None:
            if isinstance(domains, int):
                if levels is None:
                    doms = np.cumsum(domains**np.arange(16))
                    levels = np.where((N//2-diagonals)/doms <= maxs)[0][0]
                levels = max(1, levels)
                self.D = np.full(levels, domains, dtype=int)
            else:
                domains = np.atleast_1d(domains)
                levels = len(domains)
                self.D = domains
        else:
            if levels is None:
                domains = 2
                doms = np.cumsum(domains**np.arange(20))
                levels = np.where((N//2-diagonals)/doms <= maxs)[0][0]
                levels = max(1, levels)
            else:
                for domains in range(2, 100):
                    Nb = np.cumsum(domains**np.arange(levels+1))[levels]
                    s = (N//2-diagonals)//Nb
                    if s <= maxs:
                        break
            self.D = np.full(levels, domains, dtype=int)

        self.L = max(1, levels)
        Nb = get_number_of_blocks(self.L, self.D)
        self.diags = diagonals
        self.axis = 0
        # Create new (even) N with minimal size according to diags and N
        # Pad input array with zeros
        s, rest = divmod(N//2-diagonals, Nb)
        Nn = N
        if rest > 0 or Nn%2 == 1:
            s += 1
            Nn = 2*(diagonals+s*Nb)

        self.Nn = Nn
        self.s = s
        fk = []
        Nk = []
        self.Mmin = np.zeros(self.L, dtype=int)
        for level in range(self.L-1, -1, -1):
            Nx = getChebyshev(level, self.D, s, diagonals, fk, Nk, l2c)
            self.Mmin[level] = Nx
        self.fk = np.hstack(fk)
        self.Nk = np.array(Nk, dtype=int)
        Mmax = self.Mmin.max()
        TT = {}
        for level in range(self.L-1, -1, -1):
            if self.D[level] not in TT:
                TT[self.D[level]] = conversionmatrix(self.D[level], Mmax)
        self.Th = np.concatenate([TT[d] for d in np.unique(self.D)])
        self.ThT = np.array([copy(self.Th[i].transpose()) for i in range(self.Th.shape[0])])
        self.T = np.zeros((2, s, self.Mmin[-1]))
        Ti = n_cheb.chebvander(np.linspace(-1, 1, 2*s+1)[:-1], self.Mmin[-1]-1)
        self.T[0] = Ti[::2]
        self.T[1] = Ti[1::2]

    def get_M_shapes(self):
        M = []
        for level in range(self.L):
            Nb = get_number_of_blocks(level, self.D)
            M.append([np.sum((self.Nk[level][1:])**2), Nb*self.D[level]*(self.D[level]+1)//2])
        return M

    def plan(self, shape, dtype, axis, output_array=None, use_direct=-1):
        if self._output_array.shape == shape and axis == self.axis and self._output_array.dtype == dtype:
            return
        self.axis = axis
        self.sl = slicedict(axis=axis, dimensions=len(shape))
        self.si = islicedict(axis=axis, dimensions=len(shape))
        if shape[axis] > use_direct:
            oddevenshape = list(shape)
            oddevenshape[axis] = self.Nn//2
            self.cont_input_array = np.zeros(oddevenshape, dtype=dtype)
            self.cont_output_array = np.zeros(oddevenshape, dtype=dtype)
        self._output_array = np.zeros(shape, dtype=dtype) if output_array is None else output_array

    def plotrank(self):
        import matplotlib.pyplot as plt
        z = np.zeros((self.Nn, self.Nn))
        ik = 0
        for level in range(self.L-1, -1, -1):
            for block in range(get_number_of_blocks(level, self.D)):
                h = 2 * self.s * get_h(level, self.D)
                i0, j0 = get_ij(level, block, self.s, self.D, self.diags)
                for j in range(self.D[level]):
                    for i in range(j + 1):
                        z[2*i0+i*h:2*i0+(i+1)*h, 2*j0+(j+1)*h:2*j0+(j+2)*h] = self.Nk[ik]
                        ik += 1
        plt.figure()
        plt.imshow(z, cmap='gray')
        plt.title('Submatrix rank')
        plt.colorbar()
        plt.show()

def get_ij(level, block, s, D, diagonals):
    i0 = block*D[level]*s*get_h(level, D)
    j0 = diagonals+i0
    for i in range(level+1, len(D)):
        j0 += s*get_h(i, D)
    return (i0, j0)

def get_h(level, D):
    #return (self.domains-1)**(self.L-1-level) # const D
    return np.prod(D[level+1:])

def get_number_of_blocks(level, D):
    #return np.cumsum(D[0]**np.arange(level+1))[-1] # const D
    return 1+np.sum(np.cumprod(np.flip(D[:level])))

def get_number_of_submatrices(D):
    Ns = 0
    for level in range(len(D)):
        Ns += get_number_of_blocks(level, D)*D[level]*(D[level]+1)//2
    return Ns

class FMMLeg2Cheb(FMMLevel):
    """Transform Legendre coefficients to Chebyshev coefficients

    Parameters
    ----------
    input : int or input array
        The length of the array to transform or the array itself
    diagonals : int
        The number of neglected diagonals, that are treated using a
        direct approach
    domains : None, int or sequence of ints
        The domain sizes for all levels

        If domains=None then an appropriate domain size is computed
        according to the given number of levels and maxs. If the
        number of levels is also None, then domains is set to 2
        and the number of levels is computed from maxs.

        If domains is an integer, then this integer is used for each
        level, and the number of levels is either given or computed
        according to maxs

        If domains is a sequence of integers, then these are the
        domain sizes for all the levels and the length of this
        sequence is the number of levels.

    levels : None or int
        The number of levels in the hierarchical matrix

        If levels is None, then it is computed according to domains
        and maxs

    l2c : bool
        If True, the transform goes from Legendre to Chebyshev, and
        vice versa if False

    maxs : int
        The maximum size of the smallest submatrices (on the highest
        level). This number is used if the number of levels or domain
        sizes need to be computed.

    axis : int
        The axis over which to apply the transform if the input array
        is a multidimensional array.

    use_direct : int
        Use direct method if N is smaller than this number

    """
    def __init__(self, input, output_array=None, diagonals=16, domains=None, levels=None, maxs=100, axis=0, use_direct=-1):
        if isinstance(input, int):
            N = input
            shape = (N,)
            dtype = float
        elif isinstance(input, np.ndarray):
            N = input.shape[axis]
            shape = input.shape
            dtype = input.dtype
        FMMLevel.__init__(self, N, domains=domains, diagonals=diagonals, levels=levels, maxs=maxs, use_direct=use_direct)
        self.a = Omega(np.arange(self.Nn, dtype=float))
        self.plan(shape, dtype, axis, output_array, use_direct)

    def __call__(self, input_array, output_array=None, transpose=False):
        """Execute transform

        Parameters
        ----------
        input_array : array
            The array to be transformed
        output_array : None or array
            The return array. Will be created if None.
        transpose : bool
            If True, then apply the transpose operation :math:`A^Tu`
            instead of the default :math:`Au`, where :math:`A` is the
            matrix and :math:`u` is the array of Legendre coefficients.

        """
        self.plan(input_array.shape, input_array.dtype, self.axis, output_array, self.use_direct)
        if input_array.shape[self.axis] <= self.use_direct and input_array.ndim == 1:
            self._output_array[...] = 0
            _leg2cheb(input_array, self._output_array, self.a, transpose)
            if output_array is not None:
                output_array[...] = self._output_array
                return output_array
            return self._output_array
        if transpose is True:
            input_array = input_array*2/np.pi # makes copy (preserve input)
            input_array[self.si[0]] *= 0.5
        if input_array.shape[self.axis] <= self.use_direct:
            self._output_array.fill(0)
            FMMdirect1(input_array, self._output_array, self.axis, self.a, self.N//2, transpose)
        else:
            self.apply(input_array, self._output_array, transpose)
        if transpose is False:
            self._output_array *= (2/np.pi)
            self._output_array[self.si[0]] *= 0.5
        return self._output_array

    def apply(self, input_array, output_array, transpose):
        FMMcheb(input_array, output_array, self.axis, self.Nn, self.fk, self.Nk, self.T, self.Th, self.ThT, self.D, self.Mmin, self.s, self.diags, transpose)
        FMMdirect1(input_array, output_array, self.axis, self.a, self.diags, transpose)
        FMMdirect2(input_array, output_array, self.axis, self.a, 2*self.s, get_number_of_blocks(self.L, self.D), 2*self.diags, transpose)


class FMMCheb2Leg(FMMLevel):
    """Transform Chebyshev coefficients to Legendre coefficients

    Parameters
    ----------
    input : int or input array
        The length of the array to transform or the array itself
    diagonals : int
        The number of neglected diagonals, that are treated using a
        direct approach
    domains : None, int or sequence of ints
        The domain sizes for all levels

        If domains=None then an appropriate domain size is computed
        according to the given number of levels and maxs. If the
        number of levels is also None, then domains is set to 2
        and the number of levels is computed from maxs.

        If domains is an integer, then this integer is used for each
        level, and the number of levels is either given or computed
        according to maxs

        If domains is a sequence of integers, then these are the
        domain sizes for all the levels and the length of this
        sequence is the number of levels.

    levels : None or int
        The number of levels in the hierarchical matrix

        If levels is None, then it is computed according to domains
        and maxs

    l2c : bool
        If True, the transform goes from Legendre to Chebyshev, and
        vice versa if False

    maxs : int
        The maximum size of the smallest submatrices (on the highest
        level). This number is used if the number of levels or domain
        sizes need to be computed.

    axis : int
        The axis over which to apply the transform if the input array
        is a multidimensional array.

    use_direct : int
        Use direct method if N is smaller than this number
    """
    def __init__(self, input, output_array=None, diagonals=16, domains=None, levels=None, maxs=100, axis=0, use_direct=-1):
        if isinstance(input, int):
            N = input
            shape = (N,)
            dtype = float
        elif isinstance(input, np.ndarray):
            N = input.shape[axis]
            shape = input.shape
            dtype = input.dtype
        FMMLevel.__init__(self, N, domains=domains,
                          diagonals=diagonals, levels=levels, maxs=maxs, l2c=False, use_direct=use_direct)
        k = np.arange(self.Nn, dtype='d')
        k[0] = 1
        self.dn = Omega((k[::2]-2)/2)/k[::2]
        self.a = 1/(2*Omega(k)*k*(k+0.5))
        self.a[0] = 2/np.sqrt(np.pi)
        self.plan(shape, dtype, axis, output_array, use_direct)

    def __call__(self, input_array, output_array=None):
        """Execute transform

        Parameters
        ----------
        input_array : array
            The array to be transformed
        output_array : None or array
            The return array. Will be created if None.

        """
        assert isinstance(input_array, np.ndarray)
        self.plan(input_array.shape, input_array.dtype, self.axis, output_array, self.use_direct)
        if input_array.shape[self.axis] <= self.use_direct and input_array.ndim == 1:
            self._output_array[...] = 0
            _cheb2leg(input_array, self._output_array, self.dn, self.a)
            if output_array is not None:
                output_array[...] = self._output_array
                return output_array
            return self._output_array

        self.sl = sl = slicedict(axis=self.axis, dimensions=input_array.ndim)
        si = [None]*input_array.ndim
        si[self.axis] = slice(None)
        w0 = input_array.copy()
        w0[sl[slice(1, self.N)]] *= np.arange(1, self.N)[tuple(si)]

        if input_array.shape[self.axis] <= self.use_direct:
            self._output_array.fill(0)
            FMMdirect4(w0, self._output_array, self.axis, self.dn[:self.N//2], self.a[:self.N], self.N//2)
        else:
            self.apply(w0, self._output_array)

        self._output_array *= (np.arange(self.N)+0.5)[tuple(si)]
        if output_array is not None:
            output_array[...] = self._output_array
            return output_array
        return self._output_array

    def apply(self, input_array, output_array):
        FMMcheb(input_array, output_array, self.axis, self.Nn, self.fk, self.Nk, self.T, self.Th, self.ThT, self.D, self.Mmin, self.s, self.diags, False)
        FMMdirect4(input_array, output_array, self.axis, self.dn[:self.N//2], self.a[:self.N], self.diags)
        FMMdirect3(input_array, output_array, self.axis, self.dn, self.a, 2*self.s, get_number_of_blocks(self.L, self.D), 2*self.diags)

@runtimeoptimizer
def _cheb2leg(u, v, dn, a):
    pass

@runtimeoptimizer
def _leg2cheb(u, v, a, trans):
    pass

@runtimeoptimizer
def FMMdirect1(u, v, axis, a, n0, trans):
    N = u.shape[0]
    if trans is False:
        for n in range(0, n0):
            v[:(N-2*n)] += a[n]*a[n:(N-n)]*u[2*n:]
    else:
        for n in range(0, n0):
            v[2*n:] += a[n]*a[n:(N-n)]*u[:(N-2*n)]

@runtimeoptimizer
def FMMdirect2(u, v, axis, a, h, Nb, n0, trans):
    N = v.shape[0]
    for k in range(Nb):
        i0 = k*h
        j0 = n0+i0
        for n in range(0, h, 2):
            a0 = i0+(n0+n)//2
            Nm = min(N-(j0+n), h-n)
            if trans is True:
                v[j0+n:(j0+Nm)] += a[(n+n0)//2]*a[a0:a0+Nm]*u[i0:i0+Nm]
            else:
                v[i0:(i0+Nm)] += a[(n+n0)//2]*a[a0:a0+Nm]*u[j0+n:j0+n+Nm]

#from shenfun import optimization
#FMMdirect2 = optimization.numba.transforms.FMMdirect2
#FMMdirect3 = optimization.numba.transforms.FMMdirect3

@runtimeoptimizer
def FMMdirect3(u, v, axis, dn, a, h, Nb, n0):
    N = v.shape[0]
    for k in range(Nb):
        i0 = k*h
        j0 = n0+i0
        for n in range(0, h, 2):
            a0 = i0+(n0+n)//2
            Nm = min(N-(j0+n), h-n)
            v[i0:(i0+Nm)] -= dn[(n+n0)//2]*a[a0:a0+Nm]*u[j0+n:j0+n+Nm]

@runtimeoptimizer
def FMMdirect4(u, v, axis, dn, a, n0):
    N = u.shape[0]
    v[:] += np.sqrt(np.pi)*a*u
    for n in range(1, n0):
        v[:(N-2*n)] -= dn[n]*a[n:(N-n)]*u[2*n:]

@runtimeoptimizer
def FMMcheb(input_array, output_array, axis, Nn, fk, Nk, T, Th, ThT, D, Mmin, s, diags, transpose):
    assert input_array.ndim == 1, 'Use Cython for multidimensional'
    L = len(D)
    Mmax = max(Mmin)
    N = input_array.shape[0]
    cia = np.zeros(Nn//2, dtype=input_array.dtype)
    coa = np.zeros(Nn//2, dtype=input_array.dtype)
    uD = np.hstack((0, np.unique(D)))
    output_array[:] = 0
    wk = [None]*L
    ck = [None]*L
    for level in range(L):
        wk[level] = np.zeros((get_number_of_blocks(level, D), D[level], Mmax))
        ck[level] = np.zeros((get_number_of_blocks(level, D), D[level], Mmax))
    for sigma in (0, 1):
        cia[:] = 0
        coa[:] = 0
        cia[:N//2+(N%2)*(1-sigma)] = input_array[sigma::2]
        if sigma == 1:
            for level in range(L):
                wk[level][:] = 0
                ck[level][:] = 0

        ik = 0
        Nc = 0
        if transpose is False:
            for level in range(L-1, -1, -1):
                h = s*get_h(level, D)
                M0 = Mmin[level]
                s0 = np.searchsorted(uD, D[level])-1
                s1 = np.cumsum(uD)[s0]
                TT = Th[s1:s1+D[level]]
                #if level == L-1:
                #    wk[level][...] = np.dot(cia[diags+s:].reshape((D[L-1]*get_number_of_blocks(L-1, D), s)), T[sigma]).reshape((get_number_of_blocks(L-1, D), D[L-1], Mmax))

                for block in range(get_number_of_blocks(level, D)):
                    i0, j0 = get_ij(level, block, s, D, diags)
                    b0, q0 = divmod(block-1, D[level-1])
                    for q in range(D[level]):
                        if level == L-1:
                            wk[level][block, q, :M0] = np.dot(cia[j0+(q+1)*h:j0+(q+2)*h], T[sigma, :, :M0])
                        if level > 0 and block > 0:
                            wk[level-1][b0, q0, :M0] += np.dot(TT[q, :M0, :M0], wk[level][block, q, :M0])
                        for p in range(q+1):
                            M = Nk[ik]
                            ck[level][block, p, :M] += np.dot(fk[Nc:(Nc+M*M)].reshape(M, M), wk[level][block, q, :M])
                            Nc += M*M
                            ik += 1

            for level in range(L):
                M0 = Mmin[level]
                if level < L-1:
                    s0 = np.searchsorted(uD, D[level+1])-1
                    s1 = np.cumsum(uD)[s0]
                    TT = ThT[s1:s1+D[level+1]]
                    for block in range(get_number_of_blocks(level+1, D)-1):
                        b0, p0 = divmod(block, D[level])
                        for p in range(D[level+1]):
                            ck[level+1][block, p, :M0] += np.dot(TT[p, :M0, :M0], ck[level][b0, p0, :M0])

                else:
                    for block in range(get_number_of_blocks(level, D)):
                        i0, j0 = get_ij(level, block, s, D, diags)
                        for p in range(D[level]):
                            coa[i0+p*s:i0+(p+1)*s] = np.dot(T[sigma], ck[level][block, p])

        else:
            for level in range(L-1, -1, -1):
                h = s*get_h(level, D)
                M0 = Mmin[level]
                s0 = np.searchsorted(uD, D[level])-1
                s1 = np.cumsum(uD)[s0]

                for block in range(get_number_of_blocks(level, D)):
                    j0, i0 = get_ij(level, block, s, D, diags)
                    b0, q0 = divmod(block, D[level-1])
                    TT = Th[s1:s1+D[level]]
                    for p in range(D[level]):
                        for q in range(p+1):
                            if q == p:
                                if level == L-1:
                                    wk[level][block, q, :M0] = np.dot(cia[j0+q*h:j0+(q+1)*h], T[sigma, :, :M0])
                                if level > 0 and block < get_number_of_blocks(level, D)-1:
                                    wk[level-1][b0, q0, :M0] += np.dot(TT[q, :M0, :M0], wk[level][block, q, :M0])
                            M = Nk[ik]
                            ck[level][block, p, :M] += np.dot(wk[level][block, q, :M], fk[Nc:(Nc+M*M)].reshape(M, M))
                            Nc += M*M
                            ik += 1

            for level in range(L):
                M0 = Mmin[level]
                if level < L-1:
                    s0 = np.searchsorted(uD, D[level+1])-1
                    s1 = np.cumsum(uD)[s0]
                    TT = Th[s1:s1+D[level+1]]
                    for block in range(1, get_number_of_blocks(level+1, D)):
                        for p in range(D[level+1]):
                            b0, p0 = divmod(block-1, D[level])
                            ck[level+1][block, p, :M0] += np.dot(ck[level][b0, p0, :M0], TT[p, :M0, :M0])

                else:
                    for block in range(get_number_of_blocks(level, D)):
                        j0, i0 = get_ij(level, block, s, D, diags)
                        for p in range(D[level]):
                            coa[i0+(p+1)*s:i0+(p+2)*s] = np.dot(T[sigma], ck[level][block, p])

        output_array[sigma::2] = coa[:N//2+(N%2)*(1-sigma)]

def conversionmatrix(D : int, M : int) -> np.ndarray:
    from mpi4py_fft.fftw import dctn
    T = np.zeros((D, M, M))
    k = np.arange(M)
    X = np.cos((k+0.5)*np.pi/M)[None, :]
    dct = dctn(np.zeros((M, M)), axes=(1,))
    for q in range(D):
        Xk = np.cos(k[:, None]*np.arccos((X+1+2*q-D)/D))
        T[q] = dct(Xk)/M
        T[q, :, 0] /= 2
    return T



from shenfun.matrixbase import SpectralMatrix, SpectralMatDict
from shenfun.la import TDMA_O
from . import bases

CD = bases.CompactDirichlet
CN = bases.CompactNeumann
L = bases.Orthogonal


class BLLmat(SpectralMatrix):
    r"""Mass matrix :math:`B=(b_{kj}) \in \mathbb{R}^{M \times N}`, where

    .. math::

        b_{kj}=(La_j, La_k)_w,

    :math:`La_k \in` :class:`.laguerre.bases.Orthogonal` and test and trial spaces have
    dimensions of M and N, respectively.

    """
    def assemble(self, method):
        test, trial = self.testfunction, self.trialfunction
        assert isinstance(test[0], L)
        assert isinstance(trial[0], L)
        return {0: 1}

    def solve(self, b, u=None, axis=0, constraints=()):
        if u is not None:
            u[:] = b
            u /= (self.scale*self[0])
            return u

        else:
            b /= (self.scale*self[0])
            return b

    def matvec(self, v, c, format=None, axis=0):
        c[:] = v
        self.scale_array(c, self.scale*self[0])
        return c


class BCDCDmat(SpectralMatrix):
    r"""Mass matrix :math:`B=(b_{kj}) \in \mathbb{R}^{M \times N}`, where

    .. math::

        b_{kj}=(\phi_j, \phi_k)_w,

    :math:`\phi_k \in` :class:`.laguerre.bases.CompactDirichlet` and test and trial spaces have
    dimensions of M and N, respectively.

    """
    def assemble(self, method):
        test, trial = self.testfunction, self.trialfunction
        assert isinstance(test[0], CD)
        assert isinstance(trial[0], CD)
        d = {0:2., 1: -1., -1:-1.}
        return d

    def get_solver(self):
        return TDMA_O


class ACDCDmat(SpectralMatrix):
    r"""Stiffness matrix :math:`BA=(a_{kj}) \in \mathbb{R}^{M \times N}`, where

    .. math::

        a_{kj}=(\phi'_j, \phi'_k)_w,

    :math:`\phi_k \in` :class:`.laguerre.bases.CompactDirichlet` and test and trial spaces have
    dimensions of M and N, respectively.

    """
    def assemble(self, method):
        test, trial = self.testfunction, self.trialfunction
        assert isinstance(test[0], CD)
        assert isinstance(trial[0], CD)
        d = {0: 0.5,
             1: 0.25,
             -1: 0.25}
        return d

    def get_solver(self):
        return TDMA_O


mat = SpectralMatDict({
    #((CD, 0), (CD, 0)): BCDCDmat,
    #((CD, 1), (CD, 1)): ACDCDmat,
    ((L, 0), (L, 0)): BLLmat
    })



import pytest
from shenfun import FunctionSpace, TensorProductSpace, Function, np, comm, fourier


@pytest.mark.parametrize('N', ((12,)*3, (13,)*3))
def test_energy_fourier(N):
    B0 = FunctionSpace(N[0], 'F', dtype='D')
    B1 = FunctionSpace(N[1], 'F', dtype='D')
    B2 = FunctionSpace(N[2], 'F', dtype='d')
    for bases, axes in zip(((B0, B1, B2), (B0, B2, B1)),
                           ((0, 1, 2), (2, 0, 1))):
        T = TensorProductSpace(comm, bases, axes=axes)
        u_hat = Function(T)
        u_hat[:] = np.random.random(u_hat.shape) + 1j*np.random.random(u_hat.shape)
        u = u_hat.backward()
        u_hat = u.forward(u_hat)
        u = u_hat.backward(u)
        e0 = comm.allreduce(np.sum(u.v*u.v)/np.prod(N))
        e1 = fourier.energy_fourier(u_hat, T)
        assert abs(e0-e1) < 1e-10
        T.destroy()

if __name__ == '__main__':
    test_energy_fourier((12, 12, 12))



import pytest
import numpy as np
import sympy as sp
from shenfun import Function, Array, FunctionSpace, VectorSpace, \
    TensorSpace, TensorProductSpace, dot, project, comm, div, grad, config

x, y =sp.symbols('x,y', real=True)

def test_dot():
    N = 3
    D = FunctionSpace(N, 'C')
    T = TensorProductSpace(comm, (D, D))
    V = VectorSpace(T)
    S = TensorSpace(T)
    u = Function(V, buffer=(x, -y))
    bu = dot(u, u)
    assert np.linalg.norm(bu-Function(T, buffer=x**2+y**2)) < 1e-8
    gradu = project(grad(u), S)
    du = dot(gradu, u, forward_output=True)
    xy = Function(V, buffer=(x, y))
    assert np.linalg.norm(du - xy) < 1e-12
    du = dot(u, gradu, forward_output=True)
    assert np.linalg.norm(du - xy) < 1e-12
    gu = dot(gradu, gradu, forward_output=True)
    dd = Function(S, buffer=(1, 0, 0, 1))
    assert np.linalg.norm(gu - dd) < 1e-12
    T.destroy()

#@pytest.mark.skip('skipping')
def test_dot_curvilinear():
    # Define spherical coordinates without the poles
    N = 10
    basisvectors = config['basisvectors']
    config['basisvectors'] = 'normal'
    r, theta, phi = psi =sp.symbols('x,y,z', real=True, positive=True)
    rv = (r*sp.sin(theta)*sp.cos(phi), r*sp.sin(theta)*sp.sin(phi), r*sp.cos(theta))
    L0 = FunctionSpace(N, 'L', domain=(0.5, 1))
    L1 = FunctionSpace(N, 'L', domain=(0.5, np.pi/2))
    F1 = FunctionSpace(N, 'F', dtype='d')
    T = TensorProductSpace(comm, (L0, L1, F1), coordinates=(psi, rv, sp.Q.positive(sp.sin(theta))))
    V = VectorSpace(T)
    S = TensorSpace(T)
    u = Function(V, buffer=(0, sp.sin(theta), sp.cos(phi)))
    u2 = dot(u, u)
    gij = T.coors.get_metric_tensor(config['basisvectors'])
    ue = Function(T, buffer=sp.sin(theta)**2*gij[1, 1]+sp.cos(phi)**2*gij[2, 2])
    assert np.linalg.norm(ue-u2) < 1e-6
    u = Function(V, buffer=(0, r, 0))
    gu = Function(S, buffer=(0, -1, 0, 1, 0, 0, 0, 0, 1/sp.tan(theta)))
    bu = dot(gu, u)
    but = Function(V, buffer=(-r, 0, 0))
    assert np.linalg.norm(bu-but) < 1e-6
    ub = dot(u, gu)
    ubt = Function(V, buffer=(r, 0, 0))
    assert np.linalg.norm(ub-ubt) < 1e-6
    gg = dot(gu, gu)
    ggt = Function(S, buffer=(-1, 0, 0, 0, -1, 0, 0, 0, 1/sp.tan(theta)**2))
    assert np.linalg.norm(gg-ggt) < 1e-4
    T.destroy()
    config['basisvectors'] = basisvectors

if __name__ == '__main__':
    import sys
    #test_dot()
    test_dot_curvilinear()



"""
This is the **shenfun** package

What is **shenfun**?
================================

``Shenfun`` is a high performance computing platform for solving partial
differential equations (PDEs) by the spectral Galerkin method. The user
interface to shenfun is very similar to
`FEniCS <https://fenicsproject.org>`_, but applications are limited to
multidimensional tensor product grids. The code is parallelized with MPI
through the `mpi4py-fft <https://bitbucket.org/mpi4py/mpi4py-fft>`_
package.

``Shenfun`` enables fast development of efficient and accurate PDE solvers
(spectral order and accuracy), in the comfortable high-level Python language.
The spectral accuracy is ensured from using high-order *global* orthogonal
basis functions (Fourier, Legendre, Chebyshev, Laguerre, Hermite and Jacobi),
as opposed to finite element codes like `FEniCS <https://fenicsproject.org>`_
that are using low-order *local* basis functions. Efficiency is ensured
through vectorization (`Numpy <https://www.numpy.org/>`_), parallelization
(`mpi4py <https://bitbucket.org/mpi4py/mpi4py>`_) and by moving critical
routines to `Cython <https://cython.org/>`_ or
`Numba <https://numba.pydata.org>`_.

``Shenfun`` has support for solving scalar and vector equations in curvilinear
coordinates, like polar or spherical coordinates.

"""
#pylint: disable=wildcard-import,no-name-in-module

__version__ = '4.1.1'
__author__ = 'Mikael Mortensen'

import numpy as np
from mpi4py import MPI
from .config import config, dumpconfig
from . import chebyshev
from . import chebyshevu
from . import legendre
from . import laguerre
from . import hermite
from . import fourier
from . import jacobi
from . import ultraspherical
from . import matrixbase
from . import la
from .coordinates import Coordinates
from .fourier import energy_fourier
from .io import *
from .matrixbase import *
from .spectralbase import inner_product, MixedFunctionSpace, BoundaryConditions
from .forms import *
from .tensorproductspace import *
from .utilities import *
from .utilities.lagrangian_particles import *
from .utilities.integrators import *
comm = MPI.COMM_WORLD
